import json
from collections import UserDict
from typing import Callable, Self, Any, Iterable, Mapping

from pystrict import strict

from oldaplib.src.helpers.serializer import serializer


#@strict
@serializer
class ObservableDict(UserDict):
    __on_change: Callable[[Self], None]

    def __init__(self,
                 obj: Iterable | Mapping | None = None, *,
                 on_change: Callable[[Self], None] | None = None,
                 **kwargs):
        self.__on_change = on_change
        if obj:
            super().__init__(obj, **kwargs)
        else:
            super().__init__(**kwargs)

    def __setitem__(self, key, value):
        if self.__on_change:
            self.__on_change(self.copy())
        super().__setitem__(key, value)

    def __delitem__(self, key):
        if self.__on_change:
            self.__on_change(self.copy())
        super().__delitem__(key)


    def copy(self) -> Self:
        return ObservableDict(self.data.copy())

    def set_on_change(self, on_change: Callable[[Self], None]) -> None:
        self.__on_change = on_change

    def _as_dict(self):
        return self.data

    def changeset_clear(self):
        pass




import re
from typing import Self

from pystrict import strict

from oldaplib.src.enums.language import Language
from oldaplib.src.enums.xsd_datatypes import XsdValidator, XsdDatatypes
from oldaplib.src.helpers.oldaperror import OldapErrorValue
from oldaplib.src.helpers.serializer import serializer
from oldaplib.src.xsd.xsd import Xsd


#@strict
@serializer
class Xsd_language(Xsd):
    """
    Implements the [xsd:language](https://www.w3.org/TR/xmlschema11-2/#language) datatype,
    which is used to define human language tags
    conforming to [IETF BCP 47](https://en.wikipedia.org/wiki/IETF_language_tag).
    It supports a format based on RFC 4646 language tags,
    typically composed of primary language subtags optionally followed by subtags
    denoting country, region, or variant.
    """
    __value: str

    def __init__(self, value: Self | Language | str, validate: bool = True):
        """
        Constructs a new Xsd_language instance. It checks if the syntax is valid, but it does
        not test if the language tag is a valid language!
        :param value: Either a Xsd_language or a Language instance, or a string representing the language
        short name.
        :type value: Xsd_language | Language | str
        :raises ValueError: If the syntax is invalid.
        """
        if isinstance(value, Xsd_language):
            self.__value = value.__value
        elif isinstance(value, Language):
            self.__value = value.name.lower()
        else:
            if validate:
                if not XsdValidator.validate(XsdDatatypes.language, value):
                    raise OldapErrorValue(f'Invalid string "{value}" for xsd:language.')
                if not re.match('^[a-zA-Z]{2}(-[a-zA-Z]{2})?$', value):
                    raise OldapErrorValue(f'Invalid string "{value}" for xsd:language.')
            self.__value = value

    def __str__(self):
        """
        Returns the string representation of the Xsd_language instance.
        :return:
        """
        return self.__value

    def __repr__(self):
        """
        Returns the constructor string representation of the Xsd_language instance.
        :return:
        """
        return f'Xsd_language("{self.__value}")'

    def __eq__(self, other: Self | str | None) -> bool:
        """
        Checks if the Xsd_language instance is equal to the other Xsd_language instance.
        :param other: A Xsd_language instance.
        :return: True or False
        :rtype: bool
        :raises OldapErrorValue: If the syntax is invalid.
        """
        if other is None:
            return False
        if isinstance(other, Xsd_language):
            return self.__value == other.__value
        if isinstance(other, Language):
            other = other.name.lower()
        else:
            other = Xsd_language(other)
        return self.__value == other

    def __hash__(self) -> int:
        """
        Returns the hash of the Xsd_language instance.
        :return: Hash value
        :rtype: int
        """
        return hash(self.__value)

    def _as_dict(self) -> dict[str, str]:
        """
        Internal method used for JSON serialization (@serializer decorator)
        :return: dict
        """
        return {'value': self.__value}

    @property
    def toRdf(self) -> str:
        """
        Returns the Xsd_language instance as a RDF string.
        :return: RDF string
        :rtype: str
        """
        return f'"{self.__value}"^^xsd:language'




import json
from typing import Iterator, Self, Iterable

from pystrict import strict

from oldaplib.src.dtypes.rdfset import RdfSet
from oldaplib.src.enums.language import Language
from oldaplib.src.helpers.Notify import Notify
from oldaplib.src.helpers.oldaperror import OldapErrorValue, OldapErrorType, OldapErrorKey
from oldaplib.src.helpers.serializer import serializer
from oldaplib.src.xsd.xsd_string import Xsd_string


@serializer
class LanguageIn(RdfSet[Language], Notify):
    """
    This class implements the SHACL sh:languageIn datatype. It completely validates the input.
    If the validations fail, an OldapErrorValue is raised.
    """

    def __init__(self,
                 *args: Self | set[Language | str] | list[Language | str] | tuple[Language | str] | Language | str,
                 value: Self | set[Language | str] | list[Language | str] | tuple[Language | str] | Language | str | None = None):
        """
        Implementation of the SHACL sh:languageIn datatype. It completely validates the input.
        :param args: languages to be included
        :type args: Self | set[Language | str] | list[Language | str] | tuple[Language | str] | Language | str
        :param value: Internal parameter used for serialization/deserialization
        :type value: Self | set[Language | str] | list[Language | str] | tuple[Language | str] | Language | str | None
        :raises OldapErrorType: Iterable contains element that is not an instance of "Language"
        :raises OldapErrorKey: Language not existing
        """
        nargs = tuple()
        nvalue = None
        if len(args) == 0:
            if value is None:
                pass
            else:
                if isinstance(value, LanguageIn):
                    nvalue = value
                elif isinstance(value, (set, list, tuple)):
                    for v in value:
                        if not isinstance(v, (Language, str)):
                            raise OldapErrorType(f'Iterable contains element that is not an instance of "Language", but "{type(v).__name__}".')
                    try:
                        nvalue = [x if isinstance(x, Language) else Language[x.upper()] for x in value]
                    except KeyError as err:
                        raise OldapErrorKey(str(err))
                else:
                    if not isinstance(value, (Language, str)):
                        raise OldapErrorType(f'Value is not an instance of "Language", but "{type(value).__name__}".')
                    try:
                        nvalue = value if isinstance(value, Language) else Language[value.upper()]
                    except KeyError as err:
                        raise OldapErrorKey(str(err))
        elif len(args) == 1:
            if isinstance(args[0], LanguageIn):
                nargs = args
            if isinstance(args[0], (set, list, tuple)):
                for v in args[0]:
                    if not isinstance(v, (Language, str)):
                        raise OldapErrorType(f'Iterable contains element that is not an instance of "Xsd", but "{type(v).__name__}".')
                    try:
                        nargs = tuple([x if isinstance(x, Language) else Language[x.upper()] for x in args[0]])
                    except KeyError as err:
                        raise OldapErrorKey(str(err))
            else:
                if not isinstance(args[0], (LanguageIn, Language, str)):
                    raise OldapErrorType(f'Value is not an instance of "Language", but "{type(args[0]).__name__}".')
                if isinstance(args[0], LanguageIn):
                    nargs = args
                else:
                    try:
                        nargs = args if isinstance(args[0], Language) else (Language[args[0].upper()],)
                    except KeyError as err:
                        raise OldapErrorKey(str(err))
        else:
            for v in args:
                if not isinstance(v, (Language, str)):
                    raise OldapErrorType(f'Iterable contains element that is not an instance of "Xsd", but "{type(v).__name__}".')
                try:
                    nargs = tuple([x if isinstance(x, Language) else Language[x.upper()] for x in args])
                except KeyError as err:
                    raise OldapErrorKey(str(err))

        super().__init__(*nargs, value=nvalue)

    def __contains__(self, val: Language | str) -> bool:
        """
        Tests if the given language is contained in the given language set.
        :param val: Language to be tested
        :type val: Language | str
        :return: True if the given language is contained in the given language set.
        :rtype: bool
        """
        if not isinstance(val, Language):
            val = Language[str(val).upper()]
        return super().__contains__(val)

    def add(self, language: Language | Xsd_string | str) -> None:
        """
        Add a language to the given language set.
        :param language: The Language to be added to the given language set.
        :type language: Language | Xsd_string | str
        :return: None
        """
        if not isinstance(language, Language):
            try:
                language = Language[str(language).upper()]
            except ValueError as err:
                raise OldapErrorValue(str(err))
            except KeyError as err:
                raise OldapErrorKey(str(err))
        super().add(language)

    def discard(self, language: Language | Xsd_string | str) -> None:
        """
        Remove a language from the given language set.
        :param language: The Language to be removed from the given language set.
        :return: None
        :raises OldapErrorValue: If the given language is not contained in the given language set.
        """
        if not isinstance(language, Language):
            try:
                language = Language[str(language).upper()]
            except ValueError as err:
                raise OldapErrorValue(str(err))
        super().discard(language)




from typing import Any

from pystrict import strict
from rdflib import BNode

from oldaplib.src.helpers.oldaperror import OldapErrorValue
from oldaplib.src.helpers.serializer import serializer
from oldaplib.src.xsd.xsd_qname import Xsd_QName


#@strict
@serializer
class BNode(Xsd_QName):
    """
    # BNode

    Represents a blank node in the triple store. The class has the following methods:

    - *Constructor*: Initialize a blank node
    - *str()*: Return the name of the blank node
    - *repr()*: Return the Python representation of the blank node
    - *==*: Test for equality of 2 blank nodes
    - *!=*: Test for inequality of 2 blank nodes
    - *hash()*: Return the hash of the blank node
    - *value()*: Return the value of the blank node (same as str())
    """

    def __init__(self, value: Xsd_QName | str, validate: bool = True) -> None:
        """
        Construct a blank node from its name
        :param value: Name/id of the blank node
        :type value: str
        """
        super().__init__(value, validate=validate)
        if self.prefix != '_':
            raise OldapErrorValue('BNode prefix is not "_"')

    def __repr__(self) -> str:
        """
        Return the Python representation of the BNode
        :return: Python representation of the BNode
        """
        return f'BNode("{self._value}")'



from enum import unique, Enum

from oldaplib.src.enums.attributeclass import AttributeClass
from oldaplib.src.helpers.langstring import LangString
from oldaplib.src.xsd.xsd_ncname import Xsd_NCName


@unique
class OldapListAttr(AttributeClass):
    """
    This enum class represents the fields used in the project model
    """
    # order: (QName, mandatory, immutable, datatype)
    OLDAPLIST_ID = ('oldap:oldapListId', True, True, Xsd_NCName)  # virtual property, represents the RDF subject
    PREF_LABEL = ('skos:prefLabel', False, False, LangString)
    DEFINITION = ('skos:definition', False, False, LangString)



from typing import Self, Optional, Any

from pystrict import strict

from oldaplib.src.helpers.oldaperror import OldapErrorValue
from oldaplib.src.helpers.serializer import serializer
from oldaplib.src.xsd.xsd import Xsd
from oldaplib.src.xsd.xsd_ncname import Xsd_NCName


#@strict
@serializer
class Xsd_QName(Xsd):
    """
    Implements a XML Schema qualified name [xsd:QName](https://www.w3.org/TR/xmlschema-2/#QName).

    A QName consists of a prefix (itelf a NCName) and a fragment (also itself a NCName) seperatet
    by a colon (":").
    The following methods are implemented:

    - *Constructor*: Construct a QName from a QName, string (with a ":") or a prefix/fragment pair
    - *len()*: Return the length of the QName, that is the number of characters of the string representation
    - *repr()*: Return the Python representation of the QName
    - *str()*: Return the string representation of the QName
    - *==*: Test for equality
    - *!=*: Test for inequality
    - *hash()*: Return the hash of the QName
    - *prefix*: Property for the prefix of the QName
    - *fragment*: Property for the fragment of the QName

    """
    _value: str

    def __init__(self, value: Self | str | Xsd_NCName, fragment: str | Xsd_NCName | None = None, validate: bool = True) -> None:
        """
        Construct a QName from a QName, string (with a ":") or a prefix/fragment pair
        :param value: A Qname, string (with a ":") or a prefix as NCName or string
        :type value: Xsd_QName | str | Xsd_NCName
        :param fragment: A NCName or string (conforming to NCName the convention) for the fragment part [optinonal]
        :type fragment: str | Xsd_NCName | None
        :raises OldapErrorValue: If the QName representation is invalid
        """
        if fragment is None:
            if isinstance(value, Xsd_QName):
                self._value = str(value)
            elif isinstance(value, str):
                try:
                    prefix, fragment = value.split(':')
                except ValueError as err:
                    raise OldapErrorValue(f'Invalid string "{value}" for QName')
                try:
                    prefix = Xsd_NCName(prefix, validate=validate)
                    fragment = Xsd_NCName(fragment, validate=validate)
                except OldapErrorValue as err:
                    raise OldapErrorValue(f'Invalid string "{value}" for QName. Error: {err}')
                self._value = f'{prefix}:{fragment}'
            else:
                raise OldapErrorValue(f'Invalid value for QName "{value}"')
        else:
            prefix = Xsd_NCName(value, validate=validate)
            fragment = Xsd_NCName(fragment, validate=validate)
            self._value = f'{prefix}:{fragment}'

    def __len__(self) -> int:
        """
        Return the number of characters in the QName
        :return: Length of the QName
        """
        return len(self._value)

    def __add__(self, other: Xsd_NCName | str) -> Self:
        """
        Add a NCName or a valid string to the QName
        :param other: A NCName or string (conforming to NCName the convention) for the QName
        :type other: Xsd_NCName | str
        :return: The resulting Xsd_QName object
        :rtype: Xsd_QName
        :raises OldapErrorValue: If the resulting QName representation is invalid
        """
        if isinstance(other, Xsd_NCName):
            return Xsd_QName(self._value + other.value)
        else:
            tmp = Xsd_NCName(other)
            return Xsd_QName(self._value + tmp.value)

    def __iadd__(self, other: Xsd_NCName | str) -> Self:
        """
        Add a NCName or a valid string to the QName
        :param other: A NCName or string (conforming to NCName the convention) for the QName
        :type other: Xsd_NCName | str
        :return: The resulting Xsd_QName object
        :rtype: Xsd_QName
        :raises OldapErrorValue: If the resulting QName representation is invalid
        """
        if isinstance(other, Xsd_NCName):
            self._value += other.value
        else:
            tmp = Xsd_NCName(other)
            self._value += tmp.value
        return self


    def __repr__(self):
        """
        Return the Python representation of the QName
        :return: Python representation of the QName
        :rtype: str
        """
        return f'Xsd_QName("{self._value}")'

    def __str__(self):
        """
        Return the string representation of the QName
        :return: String representation of the QName
        :rtype: str
        """
        return self._value

    # @property
    # def resUri(self) -> str:
    #     return self._value

    def __eq__(self, other: Any | None) -> bool:
        """
        Test for equality of two QNames
        :param other: Another QName/str to compare with
        :return: True of False
        :rtype: bool
        """
        if other is None:
            return False
        if isinstance(other, str):
            return self._value == other
        return isinstance(other, Xsd_QName) and self._value == other._value

    def __ne__(self, other: Any):
        """
        Test for inequality of two QNames
        :param other: Another QName/str to compare with
        :type other: Xsd_QName
        :return: True of False
        :rtype: bool
        """
        if not isinstance(other, Xsd_QName):
            return False
        return self._value != other._value

    def __hash__(self) -> int:
        """
        Return the hash of the QName
        :return: Hash of the QName
        :rtype: int
        """
        return hash(self._value)

    @property
    def toRdf(self) -> str:
        """
        Return the RDF representation of the QName
        :return: RDF representation of the QName
        :rtype: str
        """
        return self._value

    def _as_dict(self) -> dict[str, str]:
        """
        Internal method for converting a QName to a JSON dictionary (@serializer decorator)
        :return: dict
        """
        return {'value': self._value}

    @property
    def value(self) -> str:
        """
        Return the value of the QName
        :return: String representation of the QName
        :rtype: str
        """
        return self._value

    @property
    def prefix(self) -> str:
        """
        Access the prefix of the QName as property
        :return: Prefix as string
        :rtype: str
        """
        parts = self._value.split(':')
        return parts[0]

    @property
    def fragment(self) -> str:
        """
        Access the fragment as fragment of the QName as property
        :return: Fragment as string
        :rtype: str
        """
        parts = self._value.split(':')
        return parts[1]



import threading

class SingletonMeta(type):
    """
    This is a thread-safe implementation of Singleton.
    """
    _instances = {}
    _lock = threading.Lock()

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            with cls._lock:
                if cls not in cls._instances:
                    cls._instances[cls] = super().__call__(*args, **kwargs)
        return cls._instances[cls]



import math
from typing import Self

from pystrict import strict

from oldaplib.src.helpers.oldaperror import OldapErrorValue, OldapErrorType
from oldaplib.src.helpers.serializer import serializer
from oldaplib.src.xsd.xsd import Xsd


#@strict
@serializer
class FloatingPoint(Xsd):
    """
    This is the superclass of all floating point based XML Schema classes. It implements the basic
    functionality of floating point XML Schema classes. These are:

    - Constructor method `__init__`
    - Comparison methods `==`, `!=`, `<__lt__>`, `<=`, `>`, `>=`
    - Serializer methods `str(XXX)`, `repr(XXX)`,
    - RDF property `toRdf`
    - Accessor methods `XXX.value`, `float(XX)`
    - Hashing methods `hash(XXX)`

    The validation flag does not have an effect.

    """
    _value: float

    def __init__(self, value: Self | float | str, validate: bool = True):
        """
        Constructor for Floating Point
        :param value: The initial value. May not be None
        :type value: A FloatingPoint value, a float value or a string that can be interpreted as float.
        :param validate: Whether to validate the initial value
        :type validate: bool
        :raises OldapErrorValue: If the value is not a float.
        :raises OldapErrorType: If the value is not a float.
        """
        if isinstance(value, FloatingPoint):
            self._value = value._value
        elif isinstance(value, float):
            self._value = value
        else:
            try:
                self._value = float(value)
            except ValueError as err:
                raise OldapErrorValue(str(err))
            except TypeError as err:
                raise OldapErrorType(str(err))

    def __float__(self) -> float:
        """
        Returns the value as a float.
        :return: float value of instance
        :rtype: float
        """
        return self._value

    def __str__(self) -> str:
        """
        Returns the value as a string. Special numbers are "NaN", "INF" and "-INF"
        :return: Value converted to string
        :rtype: str
        """
        match str(self._value):
            case 'nan':
                return 'NaN'
            case 'inf':
                return 'INF'
            case '-inf':
                return '-INF'
            case _:
                return str(self._value)

    def __repr__(self) -> str:
        """
        Returns the value as constructor statement string
        :return: Constructor statement string
        :rtype: str
        """
        if math.isnan(self._value):
            valstr = '"NaN"'
        elif math.isinf(self._value):
            if self._value < 0.0:
                valstr = '"-INF"'
            else:
                valstr = '"INF"'
        else:
            valstr = str(self._value)
        return f'{type(self).__name__}({valstr})'

    def __eq__(self, other: Self | float | int | str | None) -> bool:
        """
        Test for equality
        :param other: The value to compare with
        :type other: Self | float | str | None
        :return: True or False
        :rtype: bool
        :raises OldapErrorValue: If the value is not comparable to a float.
        """
        if other is None:
            return False
        if isinstance(other, str):
            other = FloatingPoint(other)
        if isinstance(other, (float, int)):
            return self._value == other
        elif isinstance(other, FloatingPoint):
            return self._value == other._value
        else:
            raise OldapErrorValue(f'Cannot compare FloatingPoint("{self._value}") to {type(other)}')

    def __ne__(self, other: Self | float | int | str | None) -> bool:
        """
        Test for inequality
        :param other: The value to compare with
        :type other: Self | float | str | None
        :return: True or False
        :rtype: bool
        :raises OldapErrorValue: If the value is not comparable to a float.
        """
        if other is None:
            return True
        if isinstance(other, str):
            other = FloatingPoint(other)
        if isinstance(other, (float, int)):
            return self._value != other
        elif isinstance(other, FloatingPoint):
            return self._value != other._value
        else:
            raise OldapErrorValue(f'Cannot compare FloatingPoint("{self._value}") to {type(other).__name__}')

    def __lt__(self, other: Self | float | int | str) -> bool:
        """
        Test for less-than
        :param other: The value to compare with
        :type other: Self | float | str | None
        :return: True or False
        :rtype: bool
        :raises OldapErrorValue: If the value is not comparable to a float.
        """
        if isinstance(other, str):
            other = FloatingPoint(other)
        if isinstance(other, (float, int)):
            return self._value < other
        elif isinstance(other, FloatingPoint):
            return self._value < other._value
        else:
            raise OldapErrorValue(f'Cannot compare FloatingPoint("{self._value}") to {type(other).__name__}')

    def __le__(self, other: Self | float | int | str) -> bool:
        """
        Test for less or equal than
        :param other: The value to compare with
        :type other: Self | float | str
        :return: True or False
        :rtype: bool
        :raises OldapErrorValue: If the value is not comparable to a float.
        """
        if isinstance(other, str):
            other = FloatingPoint(other)
        if isinstance(other, (float, int)):
            return self._value <= other
        elif isinstance(other, FloatingPoint):
            return self._value <= other._value
        else:
            raise OldapErrorValue(f'Cannot compare FloatingPoint("{self._value}") to {type(other).__name__}')

    def __gt__(self, other: Self | float | int | str) -> bool:
        """
        Test for greater than
        :param other: The value to compare with
        :type other: Self | float | str
        :return: True or False
        :rtype: bool
        :raises OldapErrorValue: If the value is not comparable to a float.
        """
        if isinstance(other, str):
            other = FloatingPoint(other)
        if isinstance(other, (float, int)):
            return self._value > other
        elif isinstance(other, FloatingPoint):
            return self._value > other._value
        else:
            raise OldapErrorValue(f'Cannot compare FloatingPoint("{self._value}") to {type(other).__name__}')

    def __ge__(self, other: Self | float | int | str) -> bool:
        """
        Test for greater or equal than
        :param other: The value to compare with
        :type other: Self | float | str
        :return: True or False
        :rtype: bool
        :raises OldapErrorValue: If the value is not comparable to a float.
        """
        if isinstance(other, str):
            other = FloatingPoint(other)
        if isinstance(other, (float, int)):
            return self._value >= other
        elif isinstance(other, FloatingPoint):
            return self._value >= other._value
        else:
            raise OldapErrorValue(f'Cannot compare FloatingPoint("{self._value}") to {type(other).__name__}')

    def __hash__(self) -> int:
        """
        Returns the hash value of the instance
        :return: Hash value of instance
        :rtype: int
        """
        return hash(self._value)

    @property
    def value(self) -> float:
        """
        Returns the value of the instance
        :return: Value of instance as float
        :rtype: float
        """
        return self._value

    def _toRdf(self, xsdtype: str = 'xsd:float') -> str:
        """
        Helper for converting instances to a RDF string
        :param xsdtype: XML Schema type to use
        :type xsdtype: XML Schema datatype to use (prefix: "xsd" to be used)
        :return: string representation of the instance as RDF
        :rtype: str
        """
        if math.isnan(self):
            return '"NaN"^^' + xsdtype
        elif math.isinf(self):
            if self < 0.0:
                return '"-INF"^^' + xsdtype
            else:
                return '"INF"^^' + xsdtype
        else:
            return f'"{self}"^^' + xsdtype

    @property
    def toRdf(self) -> str:
        """
        Converts the instance to a RDF string as xsd:float
        :return: RDF string representation of the instance as RDF
        :rtype: str
        """
        return self._toRdf('xsd:float')

    def _as_dict(self) -> dict:
        """
        Used internally for JSON serialization using @serialisation decorator
        :return: Representation of the instance as dict
        :rtype: dict
        """
        return {'value': self._value}

if __name__ == '__main__':
    f = FloatingPoint("1.0")









import re
from typing import Self

from pystrict import strict

from oldaplib.src.xsd.xsd_string import Xsd_string
from oldaplib.src.helpers.oldaperror import OldapErrorValue
from oldaplib.src.helpers.serializer import serializer
from oldaplib.src.xsd.xsd import Xsd


@strict
@serializer
class Xsd_Name(Xsd):
    """
    Implements the XML Schema [xsd:Name](https://www.w3.org/TR/xmlschema11-2/#Name) datatype
    """
    __value: str

    def __init__(self, value: Self | str, validate: bool = True):
        """
        Constructs an Xsd_Name instance.
        :param value: An Xsd_Name instance or a string conforming to the XML Schema [xsd:Name]
        :raises OldapErrorValue: If an invalid value is passed.
        """
        if isinstance(value, Xsd_Name):
            self.__value = value.__value
        else:
            if validate:
                if not re.match("^[a-zA-Z_][\\w.\\-:_]*$", value):
                    raise OldapErrorValue(f'Invalid string "{value}" for xsd:Name.')
            self.__value = value

    @classmethod
    def fromRdf(cls, value: str) -> Self:
        """
        Constructs an Xsd_Name instance from a RDF string.
        :param value: An RDF string conforming to the XML Schema [xsd:Name]
        :type value: str
        :return: an Xsd_Name instance.
        :rtype: Xsd_Name
        """
        value = Xsd_string.unescaping(value)
        return cls(value, validate=False)#

    def __str__(self):
        """
        Returns the string representation of the Xsd_Name instance.
        :return: String representation of the Xsd_Name instance.
        :rtype: str
        """
        return self.__value

    def __repr__(self):
        """
        Returns the constructor string representation of the Xsd_Name instance.
        :return: Constructor string representation of the Xsd_Name instance.
        :rtype: str
        """
        return f'Xsd_Name("{self.__value}")'

    def __eq__(self, other: Self | str | None) -> bool:
        """
        Returns true if the Xsd_Name instance is equal to the other Xsd_Name instance.
        :param other: An other Xsd_Name instance.
        :return: True or False
        :rtype: bool
        :raises OldapErrorValue: If an invalid value is passed.
        """
        if other is None:
            return False
        if isinstance(other, Xsd_Name):
            return self.__value == other.__value
        else:
            return self.__value == other

    def __hash__(self) -> int:
        """
        Returns the hash value of the Xsd_Name instance.
        :return: Hash value of the Xsd_Name instance.
        :rtype: int
        """
        return hash(self.__value)

    def _as_dict(self) -> dict[str, str]:
        """
        Internal method to serialize the Xsd_Name instance to JSON (@serializer decorator).
        :return: dict
        """
        return {'value': self.__value}

    @property
    def toRdf(self) -> str:
        """
        Returns the string representation of the Xsd_Name instance.
        :return: RDF string representation of the Xsd_Name instance.
        :rtype: str
        """
        return f'"{Xsd_string.escaping(self.__value)}"^^xsd:Name'

    @property
    def value(self) -> str:
        """
        Returns the string representation of the Xsd_Name instance.
        :return: String representation of the Xsd_Name instance.
        :rtype: str
        """
        return self.__value



import sys
from argparse import ArgumentParser
from pathlib import Path

from oldaplib.src.connection import Connection
from oldaplib.src.helpers.oldaperror import OldapError
from oldaplib.src.oldaplist_helpers import load_list_from_yaml, print_sublist
from oldaplib.src.project import Project


def load_list():
    parser = ArgumentParser(prog="load_list",
                            description="Loads YAML file with hierarchical list.")
    parser.add_argument("file", help="YAML file with hierarchical list")
    parser.add_argument('-v', '--verbose', action='store_true', help="Show some informational output")
    parser.add_argument('-u', '--user', required=True, help="Username")
    parser.add_argument('-p', '--password', required=True, help="Password")
    parser.add_argument('--project', required=True, help="Project ID")
    #parser.add_argument('-h', '--help', help="Show help information")
    args = parser.parse_args()

    try:
        connection = Connection(server='http://localhost:7200',
                                repo="oldap",
                                userId=args.user,
                                credentials=args.password,
                                context_name="DEFAULT")
        project = Project.read(connection, args.project)
        path = Path(args.file)
        listnodes = load_list_from_yaml(con=connection,
                                        project=args.project,
                                        filepath=path)
        if args.verbose:
            listnode = listnodes[0]
            print_sublist(listnode.nodes)
    except OldapError as error:
        print(f'ERROR: {error}!', file=sys.stderr)
        exit(-1)
    except FileNotFoundError as error:
        print(f'ERROR: {error}!', file=sys.stderr)
        exit(-1)


if __name__ == '__main__':
    load_list()



import re
from typing import Tuple, Self

from pystrict import strict

from oldaplib.src.enums.xsd_datatypes import XsdValidator, XsdDatatypes
from oldaplib.src.helpers.oldaperror import OldapErrorValue
from oldaplib.src.helpers.serializer import serializer
from oldaplib.src.xsd.xsd import Xsd


#@strict
@serializer
class Xsd_gMonth(Xsd):
    """
    Implements the XML Schema [xsd:gMonth](https://www.w3.org/TR/xmlschema11-2/#gMonth) datatype
    """
    __month: int
    __tz: Tuple[int, int] | None
    __zulu: bool | None

    def __init__(self, value: Self | str, validate: bool = True):
        """
        Constructor of Xsd_gMonth class
        :param value: An Xsd_gMonth instance or a valid string conforming to the
        :type value: Xsd | str
        :raises OldapErrorValue: If a Xsd_gMonth instance or string is not valid
        """
        if isinstance(value, Xsd_gMonth):
            self.__month = value.__month
            self.__tz = value.__tz
            self.__zulu = value.__zulu
        else:
            if validate:
                if not XsdValidator.validate(XsdDatatypes.gMonth, value):
                    raise OldapErrorValue(f'Invalid string "{value}" for xsd:gMonth.')
            res = re.split("--([0-9]{2})((([+-][0-9]{2}):([0-9]{2}))|(Z))?", value)
            if len(res) != 8:
                raise OldapErrorValue(f'Invalid string "{value}" for xsd:gMonth.')
            self.__month = int(res[1])
            if res[2] == 'Z':
                self.__tz = (0, 0)
                self.__zulu = True
            elif res[2] is not None:
                self.__tz = (int(res[4]), int(res[5]))
                self.__zulu = False
            else:
                self.__tz = None
                self.__zulu = None
        if self.__month < 1 or self.__month > 12:
            raise OldapErrorValue(f'Invalid string "{value}" for xsd:gMonth.')

    def __str__(self):
        """
        String representation of Xsd_gMonth instance
        :return:
        """
        s = f'--{self.__month:02}'
        if self.__tz is not None:
            if self.__zulu:
                s += 'Z'
            else:
                s += f'{self.__tz[0]:0=+3}:{self.__tz[1]:02}'
        return s

    def __repr__(self):
        """
        Constructor string representation of Xsd_gMonth instance
        :return: Constructor string
        """
        return f'Xsd_gMonth("{str(self)}")'

    def __eq__(self, other: Self | str | None) -> bool:
        """
        Equality check for Xsd_gMonth instance
        :param other: Other Xsd_gMonth instance or valid gMonth string
        :type other: Xsd_gMonth | str | None
        :return: True or False
        :raises OldapErrorValue: If Xsd_gMonth instance or string is not valid
        """
        if other is None:
            return False
        if not isinstance(other, Xsd_gMonth):
            other = Xsd_gMonth(other)
        if self.__month != other.__month:
            return False
        if self.__tz != other.__tz:
            return False
        return True

    def __hash__(self) -> int:
        """
        Hashing function for Xsd_gMonth instance
        :return: Hash value
        """
        return hash(str(self))

    def _as_dict(self) -> dict[str, str]:
        """
        Internal method for serializing Xsd_gMonth instance to JSON (@serilaizer decorator)
        :return: dict
        """
        return {'value': str(self)}

    @property
    def toRdf(self) -> str:
        """
        Converts Xsd_gMonth instance to RDF string
        :return: RDF string
        """
        return f'"{str(self)}"^^xsd:gMonth'




"""
# InProject Class

The InProject class is a helper class that is used to record the per-project administrative permissions for
a particular user. It's not meant to be used without the context of a user, that is as a _property_ of a User.
"""
from dataclasses import dataclass
from enum import unique
from typing import Dict, Callable, Set, Self, ItemsView, KeysView, Iterator, Any

from pystrict import strict

from oldaplib.src.enums.action import Action
from oldaplib.src.enums.attributeclass import AttributeClass
from oldaplib.src.helpers.Notify import Notify
from oldaplib.src.helpers.attributechange import AttributeChange
from oldaplib.src.xsd.iri import Iri
from oldaplib.src.xsd.xsd_anyuri import Xsd_anyURI
from oldaplib.src.xsd.xsd_qname import Xsd_QName
from oldaplib.src.helpers.observable_set import ObservableSet
from oldaplib.src.enums.adminpermissions import AdminPermission
from oldaplib.src.helpers.oldaperror import OldapErrorValue, OldapErrorKey
from oldaplib.src.helpers.serializer import serializer
import json


#@strict
@serializer
class InProjectClass(Notify):
    """
    Implements the administrative permission a user has for the projects the user is associated with.
    """
    __setdata: dict[Iri, ObservableSet]
    _changeset: dict[Iri, AttributeChange]

    def __init__(self,
                 setdata: Self | Dict[Iri | str, set[AdminPermission | str] | ObservableSet] | None = None,
                 notifier: Callable[[Iri], None] | None = None,
                 notify_data: Iri | None = None) -> None:
        """
        Constructor of the class. The class acts like a dictionary and allows the access to the permission
        set for a project using the QName of the project as the key: ```perms = t.in_project[QName('ex:proj')]```.
        It supports the getting, setting and deleting a permission set.
        In addition, the following methods are implemented:

        - _get()_: gets the permission set or returns `None`if it doesn't exist for the given project'
        - _copy()_: Creates a deep copy of the given instance
        - _==_: Check for equality of 2 instances
        - _!=_: Check for inequality of 2 instances

        :param setdata: A dictionary with the QName/anyURI of the project as key and the set of permissions as value
        :type setdata: Dict[str | Xsd_QName, Set[AdminPermission] | ObservableSet[AdminPermission]] | None
        :param on_change: A callable that is called whenever the instance has been changed
        :type on_change: Callable[[str, ObservableSet[AdminPermission] | None], None]
        """
        super().__init__(notifier=notifier, data=notify_data)
        self.__setdata = {}
        self._changeset = {}
        if setdata is not None:
            if isinstance(setdata, InProjectClass):
                self.__setdata = setdata.__setdata
            else:
                for key, value in setdata.items():
                    key = Iri(key)
                    self.__setdata[key] = self.__perms(key, value)
        self.clear_changeset()

    def __perms(self,
                key: Iri,
                value: set[AdminPermission | str] | ObservableSet | None) -> ObservableSet:
        perms = ObservableSet(notifier=self.__on_set_changed, notify_data=key)
        if value is None:
            return perms
        for permission in value:
            if isinstance(permission, str):
                try:
                    if permission.find(':') >= 0:
                        perms.add(AdminPermission(permission))
                    else:
                        perms.add(AdminPermission('oldap:' + permission))
                except ValueError as err:
                    raise OldapErrorValue(str(err))
            elif permission in AdminPermission:
                perms.add(permission)
            else:
                raise OldapErrorValue(f'{permission} is not a valid AdminPermission')
        return perms

    def __bool__(self) -> bool:
        return bool(self.__setdata)

    def __len__(self) -> int:
        return len(self.__setdata)

    def __on_set_changed(self, key: Iri):
        if self._changeset.get(key) is None:
            self._changeset[key] = AttributeChange(self.__setdata.get(key), Action.MODIFY)
        self.notify()

    def __getitem__(self, key: Iri | str) -> ObservableSet:
        if not isinstance(key, Iri):
            key = Iri(key)
        try:
            return self.__setdata[key]
        except (KeyError, AttributeError) as err:
            raise OldapErrorKey(str(err), key)

    def __setitem__(self, key: Iri | str, value: set[AdminPermission | str] | ObservableSet | None) -> None:
        if not isinstance(key, Iri):
            key = Iri(key)
        if self.__setdata.get(key) is None:
            if self._changeset.get(key) is None:
                self._changeset[key] = AttributeChange(self.__setdata.get(key), Action.CREATE)
        else:
            if self._changeset.get(key) is None:
                self._changeset[key] = AttributeChange(self.__setdata[key], Action.REPLACE)
        if value is None:
            if self._changeset.get(key) is None:
                self._changeset[key] = AttributeChange(self.__setdata.get(key), Action.DELETE)
            del self.__setdata[key]
        else:
            self.__setdata[key] = self.__perms(key, value)
        self.notify()

    def __delitem__(self, key: Iri | str) -> None:
        if not isinstance(key, Iri):
            key = Iri(key)
        if self.__setdata.get(key) is not None:
            if self._changeset.get(key) is None:
                self._changeset[key] = AttributeChange(self.__setdata[key], Action.DELETE)
            del self.__setdata[key]
            self.notify()
        else:
            raise OldapErrorKey(f'Can\'t delete key "{key}" – does not exist')

    def __iter__(self) -> Iterator[Iri]:
        return iter(self.__setdata.items())

    def __contains__(self, key: Iri | str) -> bool:
        if not isinstance(key, Iri):
            key = Iri(key)
        return key in self.__setdata


    def __str__(self) -> str:
        s = ''
        for k, v in self.__setdata.items():
            l = [x.value for x in v]
            l.sort()
            s += f'{k} : {l}\n'
        return s

    @property
    def changeset(self) -> dict[Iri, AttributeChange]:
        return self._changeset

    def clear_changeset(self) -> None:
        """
        Clear the changeset. This method is only for internal use or debugging...
        :return: None
        """
        for key in self.__setdata.keys():
            self.__setdata[key].clear_changeset()
        self._changeset = {}

    def copy(self) -> Self:
        data_copy: dict[Iri, set[AdminPermission | str] | ObservableSet] = {}
        for key, val in self.__setdata.items():
            data_copy[key] = val
        return InProjectClass(data_copy, notifier=self._notifier, notify_data=self._notify_data)

    def __eq__(self, other: Self | None) -> bool:
        if other is None:
            return False
        if not isinstance(other, InProjectClass):
            raise OldapErrorValue(f'"Other must be an instance of InProjectClass, not {type(other)}"')
        return self.__setdata == other.__setdata

    def __ne__(self, other: Self | None) -> bool:
        if other is None:
            return True
        if not isinstance(other, InProjectClass):
            raise OldapErrorValue(f'"Other must be an instance of InProjectClass, not {type(other)}"')
        return self.__setdata != other.__setdata

    def get(self, key: Iri) -> ObservableSet | None:
        return self.__setdata.get(key)

    def items(self) -> ItemsView[Iri, ObservableSet]:
        return self.__setdata.items()

    def keys(self) -> KeysView:
        return self.__setdata.keys()

    def _as_dict(self) -> dict:
        tmp = {f'{str(key)}': value for key, value in self.__setdata.items()}
        return {'setdata': tmp}

    def add_admin_permission(self, project: Iri, permission: AdminPermission) -> None:
        pass


if __name__ == '__main__':
    in_proj = InProjectClass({Iri('oldap:HyperHamlet'): {AdminPermission.ADMIN_USERS,
                                                              AdminPermission.ADMIN_RESOURCES,
                                                              AdminPermission.ADMIN_CREATE}})
    for key, val in in_proj:
        print(key, val)
    jsonstr = json.dumps(in_proj, default=serializer.encoder_default)
    in_proj2 = json.loads(jsonstr, object_hook=serializer.decoder_hook)
    for k, v in in_proj2.items():
        print(f'{k} ({type(k)}) = {v}')


import threading
from typing import Optional

from oldaplib.src.xsd.xsd_qname import Xsd_QName


class PropertyClassSingleton(type):
    """
    The idea for this class came from "https://stackoverflow.com/questions/3615565/python-get-constructor-to-return-an-existing-object-instead-of-a-new-one".
    This class is used to create a singleton of the class.
    """
    def __call__(cls, *, property_class_iri: Optional[Xsd_QName] = None, **kwargs):
        print(".......in __call__()")
        with cls._lock:
            key = f'{property_class_iri}'
            if key not in cls._cache:
                self = cls.__new__(cls, property_class_iri=property_class_iri, **kwargs)
                print('*************************')
                cls.__init__(self, property_class_iri=property_class_iri, **kwargs)
                print("+++++++++++++++++++++++++")
                if property_class_iri is None:
                    return self
                cls._cache[key] = self
                print("---add to cache---", property_class_iri)
            else:
                print("---return from cache---", property_class_iri)
            return cls._cache[key]

    def __init__(cls, name, bases, attributes):
        cls._lock = threading.Lock()
        print("--initialization of cache---")
        super().__init__(name, bases, attributes)
        cls._cache = {}

    def delete_from_cache(cls, *, property_class_iri: Optional[Xsd_QName] = None, **kwargs):
        with cls._lock:
            del cls._cache[property_class_iri]

    def in_cache(cls, *, property_class_iri: Optional[Xsd_QName] = None) -> bool:
        with cls._lock:
            return property_class_iri in cls._cache



class Gaga(metaclass=PropertyClassSingleton):

    def __init__(self, *,
                 property_class_iri: Optional[Xsd_QName]):
        print("--initialization of gaga---")
        self.property_class_iri = property_class_iri

    def out(self):
        print(self.property_class_iri)

    @classmethod
    def factory(cls, *, property_class_iri: Optional[Xsd_QName], recache: bool = False):
        if cls.in_cache(property_class_iri=property_class_iri):
            print("*****> in Cache")
        else:
            print("*****> NOT in Cache")
        if recache:
            cls.delete_from_cache(property_class_iri=property_class_iri)
        return cls(property_class_iri=property_class_iri)

if __name__ == '__main__':
    gaga = Gaga(property_class_iri=Xsd_QName('test:gaga'))
    gaga.out()
    gugus = Gaga(property_class_iri=Xsd_QName('test:gugus'))
    gugus.out()
    test = Gaga(property_class_iri=Xsd_QName('test:gaga'))
    test2 = Gaga.factory(property_class_iri=Xsd_QName('test:gaga'))
    test3 = Gaga.factory(property_class_iri=Xsd_QName('test:gaga'), recache=True)



from pystrict import strict

from oldaplib.src.helpers.serializer import serializer
from oldaplib.src.xsd.xsd_ncname import Xsd_NCName


#@strict
@serializer
class Xsd_IDREF(Xsd_NCName):
    """
    Implements the XML Schema [xsd:IDREF](https://www.w3.org/TR/xmlschema11-2/#IDREF) datatype.
    Inherits from Xsd_NCName.
    """

    def __repr__(self):
        """
        Constrcutor string representation
        :return:
        """
        return f'Xsd_IDREF("{str(self)}")'

    @property
    def toRdf(self) -> str:
        """
        RDF representation of Xsd_IDREF
        :return:
        """
        return f'"{str(self)}"^^xsd:IDREF'



import unittest
from copy import deepcopy
from pathlib import Path
from time import sleep

from oldaplib.src.connection import Connection
from oldaplib.src.enums.userattr import UserAttr
from oldaplib.src.helpers.context import Context
from oldaplib.src.helpers.observable_set import ObservableSet
from oldaplib.src.helpers.tools import str2qname_anyiri
from oldaplib.src.xsd.iri import Iri
from oldaplib.src.xsd.xsd_anyuri import Xsd_anyURI
from oldaplib.src.xsd.xsd_boolean import Xsd_boolean
from oldaplib.src.xsd.xsd_qname import Xsd_QName
from oldaplib.src.xsd.xsd_ncname import Xsd_NCName
from oldaplib.src.helpers.oldaperror import OldapErrorNotFound, OldapErrorAlreadyExists, OldapErrorValue, OldapErrorNoPermission, OldapError
from oldaplib.src.enums.adminpermissions import AdminPermission
from oldaplib.src.user import User
from oldaplib.src.in_project import InProjectClass
from oldaplib.src.xsd.xsd_string import Xsd_string

def find_project_root(current_path):
    # Climb up the directory hierarchy and check for a marker file
    path = Path(current_path).absolute()
    while not (path / 'pyproject.toml').exists():
        if path.parent == path:
            # Root of the filesystem, file not found
            raise RuntimeError('Project root not found')
        path = path.parent
    return path


class TestUser(unittest.TestCase):
    _context: Context
    _connection: Connection
    _unpriv: Connection

    @classmethod
    def setUp(cls):
        super().setUpClass()
        project_root = find_project_root(__file__)

        cls._context = Context(name="DEFAULT")

        cls._connection = Connection(server='http://localhost:7200',
                                     repo="oldap",
                                     userId="rosenth",
                                     credentials="RioGrande",
                                     context_name="DEFAULT")
        cls._unpriv = Connection(server='http://localhost:7200',
                                 repo="oldap",
                                 userId="fornaro",
                                 credentials="RioGrande",
                                 context_name="DEFAULT")

        # user = User(con=cls._connection, userId=NCName("coyote"))
        # user.delete()
        cls._connection.clear_graph(Xsd_QName('oldap:admin'))

        file = project_root / 'oldaplib' / 'ontologies' / 'admin.trig'
        cls._connection.upload_turtle(file)
        sleep(1)  # upload may take a while...

    def tearDown(self):
        pass

    #  @unittest.skip('Work in progress')
    def test_constructor(self):
        user = User(con=self._connection,
                    userId=Xsd_NCName("testuser_without"),
                    familyName="TestWithout",
                    givenName="TestWithout",
                    email="a.b@test.com",
                    credentials="Ein@geheimes&Passw0rt")

        user = User(con=self._connection,
                    userId=Xsd_NCName("testuser_empty"),
                    familyName="TestEmpty",
                    givenName="TestEmpty",
                    email="a.b@test.com",
                    credentials="Ein@geheimes&Passw0rt",
                    inProject={})

        user = User(con=self._connection,
                    userId=Xsd_NCName("testuser_empty"),
                    familyName="TestEmpty",
                    givenName="TestEmpty",
                    email="a.b@test.com",
                    credentials="Ein@geheimes&Passw0rt",
                    inProject={Iri('oldap:HyperHamlet'): {}},
                    hasPermissions={})


        user = User(con=self._connection,
                    userId=Xsd_NCName("testuser"),
                    familyName="Test",
                    givenName="Test",
                    email="a.b@test.com",
                    credentials="Ein@geheimes&Passw0rt",
                    inProject={Iri('oldap:HyperHamlet'): {AdminPermission.ADMIN_USERS,
                                                         AdminPermission.ADMIN_RESOURCES,
                                                         AdminPermission.ADMIN_CREATE}},
                    hasPermissions={Xsd_QName('oldap:GenericView')})

        self.assertEqual(user.userId, Xsd_NCName("testuser"))
        self.assertEqual(user.familyName, "Test")
        self.assertEqual(user.givenName, "Test")
        self.assertEqual(user.email, "a.b@test.com")
        self.assertEqual(user.inProject, InProjectClass({Iri("oldap:HyperHamlet"): {
            AdminPermission.ADMIN_USERS,
            AdminPermission.ADMIN_RESOURCES,
            AdminPermission.ADMIN_CREATE
        }}))
        self.assertEqual(user.hasPermissions, {Xsd_QName('oldap:GenericView')})

    def test_user_deepcopy(self):
        user = User(con=self._connection,
                    userId=Xsd_NCName("testuser"),
                    familyName="Test",
                    givenName="Test",
                    email="a.b@test.com",
                    credentials="Ein@geheimes&Passw0rt",
                    inProject={Iri('oldap:HyperHamlet'): {AdminPermission.ADMIN_USERS,
                                                         AdminPermission.ADMIN_RESOURCES,
                                                         AdminPermission.ADMIN_CREATE}},
                    hasPermissions={Xsd_QName('oldap:GenericView')})
        user2 = deepcopy(user)
        self.assertFalse(user is user2)
        self.assertEqual(user.userId, user2.userId)
        self.assertEqual(user.userIri, user2.userIri)
        self.assertEqual(user.familyName, user2.familyName)
        self.assertEqual(user.givenName, user2.givenName)
        self.assertEqual(user.email, user2.email)
        self.assertEqual(user.credentials, user2.credentials)
        self.assertEqual(user.inProject[Iri('oldap:HyperHamlet')], user2.inProject[Iri('oldap:HyperHamlet')])
        self.assertEqual(user.hasPermissions, user2.hasPermissions)

    # @unittest.skip('Work in progress')
    def test_read_user_from_id(self):
        user = User.read(con=self._connection, userId="rosenth", ignore_cache=True)
        self.assertEqual(user.userId, Xsd_NCName("rosenth"))
        self.assertEqual(user.userIri, Iri("https://orcid.org/0000-0003-1681-4036"))
        self.assertEqual(user.familyName, Xsd_string("Rosenthaler"))
        self.assertEqual(user.givenName, Xsd_string("Lukas"))
        self.assertEqual(user.email, Xsd_string("lukas.rosenthaler@unibas.ch"))
        self.assertEqual(user.inProject, InProjectClass({
            Iri("oldap:SystemProject"): {AdminPermission.ADMIN_OLDAP},
            Iri('oldap:HyperHamlet'): {AdminPermission.ADMIN_RESOURCES}
        }))
        self.assertEqual(user.hasPermissions, {Iri("oldap:GenericRestricted"), Iri('oldap:GenericView')})

    def test_read_user_from_iri(self):
        user = User.read(con=self._connection, userId=Iri("https://orcid.org/0000-0003-1681-4036"), ignore_cache=True)
        self.assertEqual(user.userId, Xsd_NCName("rosenth"))
        self.assertEqual(user.userIri, Iri("https://orcid.org/0000-0003-1681-4036"))
        self.assertEqual(user.familyName, Xsd_string("Rosenthaler"))
        self.assertEqual(user.givenName, Xsd_string("Lukas"))
        self.assertEqual(user.email, Xsd_string("lukas.rosenthaler@unibas.ch"))
        self.assertEqual(user.inProject, InProjectClass({
            Iri("oldap:SystemProject"): {AdminPermission.ADMIN_OLDAP},
            Iri('oldap:HyperHamlet'): {AdminPermission.ADMIN_RESOURCES}
        }))
        self.assertEqual(user.hasPermissions, {Iri("oldap:GenericRestricted"), Iri('oldap:GenericView')})

    #  #unittest.skip('Work in progress')
    def test_read_unknown_user(self):
        with self.assertRaises(OldapErrorNotFound) as ex:
            user = User.read(con=self._connection, userId="nosuchuser", ignore_cache=True)
        self.assertEqual(str(ex.exception), 'User "nosuchuser" not found.')

    def test_read_user_with_cache(self):
        user = User.read(con=self._connection, userId="rosenth", ignore_cache=True)
        user2 = User.read(con=self._connection, userId="rosenth", ignore_cache=True)
        self.assertFalse(user2 is user)

        self.assertEqual(user2.userId, Xsd_NCName("rosenth"))
        self.assertEqual(user.userId, user2.userId)
        self.assertFalse(user.userId is user2.userId)

        self.assertEqual(user2.userIri, Iri("https://orcid.org/0000-0003-1681-4036"))
        self.assertEqual(user.userIri, user2.userIri)
        self.assertFalse(user.userIri is user2.userIri)

        self.assertEqual(user2.familyName, Xsd_string("Rosenthaler"))
        self.assertEqual(user.familyName, user2.familyName)
        self.assertFalse(user.familyName is user2.familyName)

        self.assertEqual(user2.givenName, Xsd_string("Lukas"))
        self.assertEqual(user.givenName, user2.givenName)
        self.assertFalse(user.givenName is user2.givenName)

        self.assertEqual(user2.email, Xsd_string("lukas.rosenthaler@unibas.ch"))
        self.assertEqual(user.email, user2.email)
        self.assertFalse(user.email is user2.email)

        self.assertEqual(user2.inProject, InProjectClass({
            Iri("oldap:SystemProject"): {AdminPermission.ADMIN_OLDAP},
            Iri('oldap:HyperHamlet'): {AdminPermission.ADMIN_RESOURCES}
        }))
        self.assertEqual(user.inProject, user2.inProject)
        self.assertFalse(user.inProject is user2.inProject)

        self.assertEqual(user2.hasPermissions, {Iri("oldap:GenericRestricted"), Iri('oldap:GenericView')})
        self.assertEqual(user.hasPermissions, user2.hasPermissions)
        self.assertFalse(user.hasPermissions is user2.hasPermissions)

        user3 = User.read(con=self._connection, userId=Iri("https://orcid.org/0000-0003-1681-4036"))
        self.assertFalse(user3 is user)

        self.assertEqual(user3.userId, Xsd_NCName("rosenth"))
        self.assertEqual(user.userId, user3.userId)
        self.assertFalse(user.userId is user3.userId)

        self.assertEqual(user3.userIri, Iri("https://orcid.org/0000-0003-1681-4036"))
        self.assertEqual(user.userIri, user3.userIri)
        self.assertFalse(user.userIri is user3.userIri)

        self.assertEqual(user3.familyName, Xsd_string("Rosenthaler"))
        self.assertEqual(user.familyName, user3.familyName)
        self.assertFalse(user.familyName is user3.familyName)

        self.assertEqual(user3.givenName, Xsd_string("Lukas"))
        self.assertEqual(user.givenName, user3.givenName)
        self.assertFalse(user.givenName is user3.givenName)

        self.assertEqual(user3.email, Xsd_string("lukas.rosenthaler@unibas.ch"))
        self.assertEqual(user.email, user3.email)
        self.assertFalse(user.email is user3.email)

        self.assertEqual(user3.inProject, InProjectClass({
            Iri("oldap:SystemProject"): {AdminPermission.ADMIN_OLDAP},
            Iri('oldap:HyperHamlet'): {AdminPermission.ADMIN_RESOURCES}
        }))
        self.assertEqual(user.inProject, user3.inProject)
        self.assertFalse(user.inProject is user3.inProject)

        self.assertEqual(user3.hasPermissions, {Iri("oldap:GenericRestricted"), Iri('oldap:GenericView')})
        self.assertEqual(user.hasPermissions, user3.hasPermissions)
        self.assertFalse(user.hasPermissions is user3.hasPermissions)


    #  #unittest.skip('Work in progress')
    def test_search_user(self):
        users = User.search(con=self._connection, userId="fornaro")
        self.assertEqual([Iri("https://orcid.org/0000-0003-1485-4923")], users)

        users = User.search(con=self._connection, familyName="Rosenthaler")
        self.assertEqual([Iri("https://orcid.org/0000-0003-1681-4036")], users)

        users = User.search(con=self._connection, givenName="John")
        self.assertEqual([Iri("urn:uuid:7e56b6c4-42e5-4a9d-94cf-d6e22577fb4b")], users)

        users = User.search(con=self._connection, inProject=Xsd_QName("oldap:HyperHamlet"))
        self.assertEqual([Iri("https://orcid.org/0000-0003-1681-4036"),
                          Iri("https://orcid.org/0000-0003-1485-4923"),
                          Iri("https://orcid.org/0000-0001-9277-3921")], users)

        users = User.search(con=self._connection, inProject=Iri("http://www.salsah.org/version/2.0/SwissBritNet"))
        self.assertEqual([Iri("https://orcid.org/0000-0002-7403-9595")], users)

        users = User.search(con=self._connection, userId="GAGA")
        self.assertEqual([], users)

    #  #unittest.skip('Work in progress')
    def test_search_user_injection(self):
        with self.assertRaises(OldapErrorValue) as ex:
            users = User.search(con=self._connection, userId="fornaro\".}\nSELECT * WHERE {?s ?p ?s})#")
        self.assertEqual(str(ex.exception), 'Invalid string "fornaro".}\nSELECT * WHERE {?s ?p ?s})#" for NCName')

        users = User.search(con=self._connection, familyName="Rosenthaler\".}\nSELECT * WHERE{?s ?p ?s})#")
        self.assertEqual(len(users), 0)
        users = User.search(con=self._connection, givenName="John\".}\nSELECT * WHERE{?s ?p ?s})#")
        self.assertEqual(len(users), 0)

        with self.assertRaises(OldapErrorValue) as ex:
            users = User.search(con=self._connection, inProject="oldap:HyperHamlet\".}\nSELECT * WHERE{?s ?p ?s})#")
        self.assertEqual(str(ex.exception), 'Invalid string for IRI: "oldap:HyperHamlet".}\nSELECT * WHERE{?s ?p ?s})#"')

    #  #unittest.skip('Work in progress')
    def test_create_user(self):
        """Test if the creation of a new user works a intended"""
        user = User(con=self._connection,
                    userIri=Iri("https://orcid.org/0000-0003-3478-9313"),
                    userId=Xsd_NCName("coyote"),
                    familyName="Coyote",
                    givenName="Wiley E.",
                    email="wiley.coyote@looneytunes.com",
                    credentials="Super-Genius",
                    inProject={Iri('oldap:HyperHamlet'): {AdminPermission.ADMIN_USERS,
                                                               AdminPermission.ADMIN_RESOURCES,
                                                               AdminPermission.ADMIN_CREATE}},
                    hasPermissions={Iri('oldap:GenericView')},
                    isActive=True)
        user.create()
        self.assertIsNotNone(user.created)
        self.assertIsNotNone(user.creator)
        self.assertIsNotNone(user.modified)
        self.assertIsNotNone(user.contributor)
        user2 = User.read(con=self._connection, userId="coyote", ignore_cache=True)
        self.assertEqual(user2.userId, user.userId)
        self.assertEqual(user2.userIri, user.userIri)
        self.assertEqual(user2.familyName, user.familyName)
        self.assertEqual(user2.givenName, user.givenName)
        self.assertEqual(user2.email, user.email)
        self.assertEqual(user2.inProject, user.inProject)
        self.assertEqual(user2.hasPermissions, user.hasPermissions)
        self.assertTrue(user2.isActive)

    #  #unittest.skip('Work in progress')
    def test_create_user_no_admin_perms(self):
        """Create a user which belongs to oldap:HyperHamlet without admin permissions"""
        user = User(con=self._connection,
                    userId=Xsd_NCName("birdy"),
                    familyName="Birdy",
                    givenName="Tweetie",
                    email="tweety.birdy@looneytunes.com",
                    credentials="Sylvester",
                    inProject={Iri('oldap:HyperHamlet'): set()},
                    hasPermissions={Iri('oldap:GenericView')},
                    isActive=True)
        user.create()
        del user
        user = User.read(con=self._connection, userId=Xsd_NCName("birdy"), ignore_cache=True)
        self.assertEqual(user.familyName, "Birdy")

    #  #unittest.skip('Work in progress')
    def test_create_user_no_in_project(self):
        """Create a user which is not associated with a project"""
        user = User(con=self._connection,
                    userId=Xsd_NCName("yogi"),
                    familyName="Baer",
                    givenName="Yogi",
                    email="yogi.baer@yellowstone.gov",
                    credentials="BuBu",
                    hasPermissions={Iri('oldap:GenericView')},
                    isActive=True)
        user.create()
        del user
        user = User.read(con=self._connection, userId=Xsd_NCName("yogi"), ignore_cache=True)
        self.assertEqual(user.familyName, "Baer")

    #  #unittest.skip('Work in progress')
    def test_create_user_no_permset(self):
        """Create a user that is associated with oldap:HyperHmalet, but is not linked a PermissionSet"""
        user = User(con=self._connection,
                    userIri=Iri("https://orcid.org/0000-0003-3478-9313"),
                    userId=Xsd_NCName("speedy"),
                    familyName="Ganzales",
                    givenName="Speedy",
                    email="speedy.gonzales@mexico.com",
                    credentials="fasterthanlight",
                    inProject={Iri('oldap:HyperHamlet'): {AdminPermission.ADMIN_USERS,
                                                               AdminPermission.ADMIN_RESOURCES,
                                                               AdminPermission.ADMIN_CREATE}},
                    isActive=True)
        user.create()
        del user
        user = User.read(con=self._connection, userId=Xsd_NCName("speedy"), ignore_cache=True)
        self.assertEqual(user.familyName, "Ganzales")


    #  #unittest.skip('Work in progress')
    def test_create_user_no_useriri(self):
        """We create a user without giving an userIri. It should recieve a URN as userIRI"""
        user = User(con=self._connection,
                    userId=Xsd_NCName("sylvester"),
                    familyName="Sylvester",
                    givenName="Cat",
                    credentials="Birdy",
                    email="sylvester.cat@looneytunes.com",
                    inProject={Iri('oldap:HyperHamlet'): {AdminPermission.ADMIN_USERS,
                                                               AdminPermission.ADMIN_RESOURCES,
                                                               AdminPermission.ADMIN_CREATE}},
                    hasPermissions={Iri('oldap:GenericView')},
                    isActive=False)
        user.create()
        del user
        user = User.read(con=self._connection, userId=Xsd_NCName("sylvester"), ignore_cache=True)
        self.assertTrue(str(user.userIri).startswith("urn:uuid:"))
        self.assertFalse(user.isActive)

    #  #unittest.skip('Work in progress')
    def test_create_user_duplicate_userid(self):
        """Test that we cannot create a duplicate userId (here "fornaro")"""
        user = User(con=self._connection,
                    userIri=Iri("https://orcid.org/0000-0003-3478-9314"),
                    userId=Xsd_NCName("fornaro"),
                    familyName="di Fornaro",
                    givenName="Petri",
                    email="petri.difornaro@italia.it",
                    credentials="Genius",
                    inProject={Iri('oldap:HyperHamlet'): {AdminPermission.ADMIN_USERS,
                                                               AdminPermission.ADMIN_RESOURCES,
                                                               AdminPermission.ADMIN_CREATE}},
                    hasPermissions={Iri('oldap:GenericView')})
        with self.assertRaises(OldapErrorAlreadyExists) as ex:
            user.create()
        self.assertEqual(str(ex.exception), 'A user with a user ID "fornaro" already exists')

    #  #unittest.skip('Work in progress')
    def test_create_user_duplicate_useriri(self):
        """Test that we cannot create a duplicate userIri"""
        user = User(con=self._connection,
                    userIri=Iri("https://orcid.org/0000-0003-1681-4036"),
                    userId=Xsd_NCName("brown"),
                    familyName="Brown",
                    givenName="Emmett",
                    email="emmet.brown@hilldale.com",
                    credentials="Time-Machine@1985",
                    inProject={Iri('oldap:HyperHamlet'): {AdminPermission.ADMIN_USERS,
                                                               AdminPermission.ADMIN_RESOURCES,
                                                               AdminPermission.ADMIN_CREATE}},
                    hasPermissions={Iri('oldap:GenericView')})
        with self.assertRaises(OldapErrorAlreadyExists) as ex:
            user.create()
        self.assertEqual(str(ex.exception), 'A user with a user IRI "https://orcid.org/0000-0003-1681-4036" already exists')

    #  #unittest.skip('Work in progress')
    def test_create_user_invalid_project(self):
        """Test that user creation will fail if we give an invalid project"""
        user = User(con=self._connection,
                    userId=Xsd_NCName("donald"),
                    familyName="Duck",
                    givenName="Donald",
                    email="donald.duck@duckville.com",
                    credentials="Entenhausen@for&Ever",
                    inProject={Iri('oldap:NotExistingproject'): {AdminPermission.ADMIN_USERS,
                                                                AdminPermission.ADMIN_RESOURCES,
                                                                AdminPermission.ADMIN_CREATE}},
                    hasPermissions={Iri('oldap:GenericView')})
        with self.assertRaises(OldapErrorValue) as ex:
            user.create()
        self.assertEqual(str(ex.exception), 'One of the projects is not existing!')

    #  #unittest.skip('Work in progress')
    def test_create_user_invalid_permset(self):
        """Test that user creation will fail if we give an invalid permission set"""
        user = User(con=self._connection,
                    userId=Xsd_NCName("donald"),
                    familyName="Duck",
                    givenName="Donald",
                    email="donald.duck@duckville.com",
                    credentials="Entenhausen@for&Ever",
                    inProject={Iri('oldap:HyperHamlet'): {AdminPermission.ADMIN_USERS,
                                                         AdminPermission.ADMIN_RESOURCES,
                                                         AdminPermission.ADMIN_CREATE}},
                    hasPermissions={Iri('oldap:GenericView'), Xsd_QName('oldap:Gaga')})
        with self.assertRaises(OldapErrorValue) as ex:
            user.create()
        self.assertEqual(str(ex.exception), 'One of the permission sets is not existing!')

    #  #unittest.skip('Work in progress')
    def test_create_user_no_privilege(self):
        """Test that user creation will fail if we do not have the permission for the given project to create users"""
        user = User(con=self._unpriv,
                    userId=Xsd_NCName("donald"),
                    familyName="Duck",
                    givenName="Donald",
                    email="donald.duck@duckville.com",
                    credentials="Entenhausen@for&Ever",
                    inProject={Iri('oldap:HyperHamlet'): {AdminPermission.ADMIN_CREATE}},
                    hasPermissions={Iri('oldap:GenericView')})
        with self.assertRaises(OldapErrorNoPermission) as ex:
            user.create()
        self.assertEqual('Actor has no ADMIN_USERS permission for project oldap:HyperHamlet', str(ex.exception), )

    #  #unittest.skip('Work in progress')
    def test_delete_user(self):
        """Test that user deletion does work"""
        user = User(con=self._connection,
                    userIri=Iri("https://orcid.org/0000-0002-9991-2055"),
                    userId=Xsd_NCName("edison"),
                    familyName="Edison",
                    givenName="Thomas A.",
                    email="thomas.aedison@ny.com",
                    credentials="Lightbulb&Phonograph",
                    inProject={Iri('oldap:HyperHamlet'): {AdminPermission.ADMIN_USERS,
                                                         AdminPermission.ADMIN_RESOURCES,
                                                         AdminPermission.ADMIN_CREATE}},
                    hasPermissions={Iri('oldap:GenericView')})
        user.create()
        user2 = User.read(con=self._connection, userId="edison", ignore_cache=True)
        self.assertEqual(user2.userIri, user.userIri)
        user2.delete()
        with self.assertRaises(OldapErrorNotFound) as ex:
            user = User.read(con=self._connection, userId="edison", ignore_cache=True)
        self.assertEqual(str(ex.exception), 'User "edison" not found.')

    #  #unittest.skip('Work in progress')
    def test_delete_user_unpriv(self):
        """Delete a user without having the permission should fail"""
        user = User.read(con=self._unpriv, userId="bugsbunny", ignore_cache=True)
        with self.assertRaises(OldapErrorNoPermission) as ex:
            user.delete()
        self.assertEqual('Actor has no ADMIN_USERS permission for project oldap:HyperHamlet', str(ex.exception))

    def test_update_user_permission_create(self):
        user = User(con=self._connection,
                    userIri=Iri("https://orcid.org/0000-0002-9991-2055"),
                    userId=Xsd_NCName("ampere"),
                    familyName="Ampère",
                    givenName="André-Marie",
                    email="ampere@current.com",
                    credentials="Current",
                    inProject={Iri('oldap:HyperHamlet'): {AdminPermission.ADMIN_USERS,
                                                         AdminPermission.ADMIN_RESOURCES,
                                                         AdminPermission.ADMIN_CREATE}})
        user.create()
        user = User.read(con=self._connection, userId="ampere", ignore_cache=True)
        user.hasPermissions = {Iri('oldap:GenericRestricted'), Iri('oldap:GenericView')}
        user.update()
        user = User.read(con=self._connection, userId="ampere", ignore_cache=True)
        self.assertEqual(user.hasPermissions, {Iri('oldap:GenericRestricted'), Iri('oldap:GenericView')})

    def test_update_user_permission_delete(self):
        user = User(con=self._connection,
                    userIri=Iri("https://orcid.org/0000-0002-9991-2055"),
                    userId=Xsd_NCName("ampere"),
                    familyName="Ampère",
                    givenName="André-Marie",
                    email="ampere@current.com",
                    credentials="Current",
                    inProject={Iri('oldap:HyperHamlet'): {AdminPermission.ADMIN_USERS,
                                                         AdminPermission.ADMIN_RESOURCES,
                                                         AdminPermission.ADMIN_CREATE}},
                    hasPermissions={Iri('oldap:GenericView')})
        user.create()
        user = User.read(con=self._connection, userId="ampere", ignore_cache=True)
        del user.hasPermissions
        user.update()
        user = User.read(con=self._connection, userId="ampere", ignore_cache=True)
        self.assertIsNone(user.hasPermissions)

    def test_update_user_permission_set_empty(self):
        user = User(con=self._connection,
                    userIri=Iri("https://orcid.org/0000-0002-9991-2055"),
                    userId=Xsd_NCName("ampere"),
                    familyName="Ampère",
                    givenName="André-Marie",
                    email="ampere@current.com",
                    credentials="Current",
                    inProject={Iri('oldap:HyperHamlet'): {AdminPermission.ADMIN_USERS,
                                                         AdminPermission.ADMIN_RESOURCES,
                                                         AdminPermission.ADMIN_CREATE}},
                    hasPermissions={Iri('oldap:GenericView')})
        user.create()
        user = User.read(con=self._connection, userId="ampere", ignore_cache=True)
        user.hasPermissions = []
        user.update()
        user = User.read(con=self._connection, userId="ampere", ignore_cache=True)
        self.assertIsNone(user.hasPermissions)

    def test_update_user_permission_set_None(self):
        user = User(con=self._connection,
                    userIri=Iri("https://orcid.org/0000-0002-9991-2055"),
                    userId=Xsd_NCName("ampere"),
                    familyName="Ampère",
                    givenName="André-Marie",
                    email="ampere@current.com",
                    credentials="Current",
                    inProject={Iri('oldap:HyperHamlet'): {AdminPermission.ADMIN_USERS,
                                                         AdminPermission.ADMIN_RESOURCES,
                                                         AdminPermission.ADMIN_CREATE}},
                    hasPermissions={Iri('oldap:GenericView')})
        user.create()
        user = User.read(con=self._connection, userId="ampere", ignore_cache=True)
        user.hasPermissions = None
        user.update()
        user = User.read(con=self._connection, userId="ampere", ignore_cache=True)
        self.assertIsNone(user.hasPermissions)

    def test_update_user_permissions_update_A(self):
        user = User(con=self._connection,
                    userIri=Iri("https://orcid.org/0000-0002-9991-2055"),
                    userId=Xsd_NCName("faraday"),
                    familyName="Faraday",
                    givenName="Michael",
                    email="michael.faraday@cage.com",
                    credentials="electromagnetism",
                    inProject={Iri('oldap:HyperHamlet'): {AdminPermission.ADMIN_USERS,
                                                         AdminPermission.ADMIN_RESOURCES,
                                                         AdminPermission.ADMIN_CREATE}},
                    hasPermissions={Iri('oldap:GenericView')})
        user.create()
        user = User.read(con=self._connection, userId="faraday", ignore_cache=True)
        user.hasPermissions.add(Iri('oldap:GenericRestricted'))
        user.update()
        user = User.read(con=self._connection, userId="faraday", ignore_cache=True)
        self.assertEqual(user.hasPermissions, {Iri('oldap:GenericRestricted'), Iri('oldap:GenericView')})

    def test_update_user_permissions_update_B(self):
        user = User(con=self._connection,
                    userIri=Iri("https://orcid.org/0000-0002-9991-2055"),
                    userId=Xsd_NCName("fourier"),
                    familyName="Fourier",
                    givenName="Joseph",
                    email="joseph.fourier@wave.com",
                    credentials="FrequencyAnalysis",
                    inProject={Iri('oldap:HyperHamlet'): {AdminPermission.ADMIN_USERS,
                                                         AdminPermission.ADMIN_RESOURCES,
                                                         AdminPermission.ADMIN_CREATE}},
                    hasPermissions={Iri('oldap:GenericRestricted'), Iri('oldap:GenericView')})
        user.create()
        user = User.read(con=self._connection, userId="fourier", ignore_cache=True)
        user.hasPermissions.remove(Iri('oldap:GenericRestricted'))
        user.update()
        user = User.read(con=self._connection, userId="fourier", ignore_cache=True)
        self.assertEqual(user.hasPermissions, {Iri('oldap:GenericView')})

    #  #unittest.skip('Work in progress')
    def test_update_user(self):
        user = User(con=self._connection,
                    userIri=Iri("https://orcid.org/0000-0002-9991-2055"),
                    userId=Xsd_NCName("edison"),
                    familyName="Edison",
                    givenName="Thomas A.",
                    email="thomas.edison@inventors.org",
                    credentials="Lightbulb&Phonograph",
                    inProject={Iri('oldap:HyperHamlet'): {AdminPermission.ADMIN_USERS,
                                                         AdminPermission.ADMIN_RESOURCES,
                                                         AdminPermission.ADMIN_CREATE}},
                    hasPermissions={Iri('oldap:GenericView')})
        user.create()
        user2 = User.read(con=self._connection, userId="edison", ignore_cache=True)
        user2.userId = "aedison"
        user2.familyName = "Edison et al."
        user2.givenName = "Thomas"
        user2.email = "th.edison@audiorec.com"

        user2.hasPermissions.add(Iri('oldap:GenericRestricted'))
        user2.hasPermissions.add(Iri('hyha:HyperHamletMember'))
        user2.hasPermissions.remove(Iri('oldap:GenericView'))
        user2.inProject[Iri('oldap:SystemProject')] = {AdminPermission.ADMIN_USERS, AdminPermission.ADMIN_RESOURCES}
        user2.inProject[Iri('oldap:HyperHamlet')].remove(AdminPermission.ADMIN_USERS)
        user2.inProject[Iri('oldap:HyperHamlet')].add(AdminPermission.ADMIN_LISTS)
        user2.update()
        user3 = User.read(con=self._connection, userId="aedison", ignore_cache=True)
        self.assertEqual(user3.familyName, "Edison et al.")
        self.assertEqual(user3.givenName, "Thomas")
        self.assertEqual(user3.email, "th.edison@audiorec.com")
        self.assertEqual({Iri('oldap:GenericRestricted'), Iri('hyha:HyperHamletMember')}, user3.hasPermissions)
        user3.hasPermissions.add(Iri('oldap:DoesNotExist'))
        with self.assertRaises(OldapErrorValue) as ex:
            user3.update()
            self.assertEqual(str(ex.exception), 'One of the permission sets is not existing!')

        self.assertEqual(user3.inProject[Iri('oldap:HyperHamlet')], {AdminPermission.ADMIN_RESOURCES,
                                                                     AdminPermission.ADMIN_CREATE,
                                                                     AdminPermission.ADMIN_LISTS})
        self.assertEqual(user3.inProject[Iri('oldap:SystemProject')], {AdminPermission.ADMIN_USERS,
                                                                       AdminPermission.ADMIN_RESOURCES})
        self.assertEqual(InProjectClass({Iri('oldap:HyperHamlet'): {AdminPermission.ADMIN_RESOURCES,
                                                                    AdminPermission.ADMIN_CREATE,
                                                                    AdminPermission.ADMIN_LISTS},
                                         Iri('oldap:SystemProject'): {AdminPermission.ADMIN_USERS,
                                                                           AdminPermission.ADMIN_RESOURCES}}), user3.inProject)
        del user3
        user4 = User.read(con=self._connection, userId="aedison", ignore_cache=True)
        user4.inProject = InProjectClass({Iri('oldap:HyperHamlet'): {AdminPermission.ADMIN_USERS}})
        user4.update()
        del user4

    def test_update_user_in_project_set(self):
        user = User(con=self._connection,
                    userIri=Iri("https://orcid.org/0000-0002-9991-2066"),
                    userId=Xsd_NCName("tesla"),
                    familyName="Tesla",
                    givenName="Nikolai",
                    email="nikolai.tesla@tesla.com",
                    credentials="Alternative current",
                    hasPermissions={Iri('oldap:GenericView')})
        user.create()
        user = User.read(con=self._connection, userId="tesla", ignore_cache=True)
        user.inProject = InProjectClass({Iri('http://www.salsah.org/version/2.0/SwissBritNet'): ObservableSet({AdminPermission.ADMIN_RESOURCES, AdminPermission.ADMIN_CREATE})})
        user.update()
        user = User.read(con=self._connection, userId="tesla", ignore_cache=True)
        self.assertEqual(user.inProject[Iri('http://www.salsah.org/version/2.0/SwissBritNet')], {AdminPermission.ADMIN_RESOURCES, AdminPermission.ADMIN_CREATE})

    def test_update_user_in_project_delete(self):
        user = User(con=self._connection,
                    userIri=Iri("https://orcid.org/0000-0002-9991-2066"),
                    userId=Xsd_NCName("tesla"),
                    familyName="Tesla",
                    givenName="Nikolai",
                    email="nikolai.tesla@tesla.com",
                    credentials="Alternative current",
                    inProject={Iri('oldap:HyperHamlet'): {AdminPermission.ADMIN_USERS,
                                                          AdminPermission.ADMIN_RESOURCES}},
                    hasPermissions={Iri('oldap:GenericView')})
        user.create()
        user = User.read(con=self._connection, userId="tesla", ignore_cache=True)
        del user.inProject
        user.update()
        user = User.read(con=self._connection, userId="tesla", ignore_cache=True)
        self.assertIsNone(user.inProject)

    def test_update_user_in_project_modify_A(self):
        user = User(con=self._connection,
                    userIri=Iri("https://orcid.org/0000-0002-9991-2066"),
                    userId=Xsd_NCName("tesla"),
                    familyName="Tesla",
                    givenName="Nikolai",
                    email="nikolai.tesla@tesla.com",
                    credentials="Alternative current",
                    inProject={Iri('oldap:HyperHamlet'): {AdminPermission.ADMIN_USERS,
                                                          AdminPermission.ADMIN_RESOURCES}},
                    hasPermissions={Iri('oldap:GenericView')})
        user.create()
        user = User.read(con=self._connection, userId="tesla", ignore_cache=True)
        user.inProject[Iri('http://www.salsah.org/version/2.0/SwissBritNet')] = {AdminPermission.ADMIN_RESOURCES}
        user.update()
        user = User.read(con=self._connection, userId="tesla", ignore_cache=True)
        self.assertEqual(user.inProject[Iri('oldap:HyperHamlet')], {AdminPermission.ADMIN_USERS, AdminPermission.ADMIN_RESOURCES})
        self.assertEqual(user.inProject[Iri('http://www.salsah.org/version/2.0/SwissBritNet')], {AdminPermission.ADMIN_RESOURCES})

    def test_update_user_in_project_modify_B(self):
        user = User(con=self._connection,
                    userIri=Iri("https://orcid.org/0000-0002-9991-2067"),
                    userId=Xsd_NCName("volta"),
                    familyName="Volta",
                    givenName="Alessandro",
                    email="alessandro.volta@tension.com",
                    credentials="ElectricTension",
                    inProject={Iri('oldap:HyperHamlet'): {AdminPermission.ADMIN_USERS,
                                                          AdminPermission.ADMIN_RESOURCES},
                               Iri('http://www.salsah.org/version/2.0/SwissBritNet'): {AdminPermission.ADMIN_MODEL}},
                    hasPermissions={Iri('oldap:GenericView')})
        user.create()
        user = User.read(con=self._connection, userId="volta", ignore_cache=True)
        user.inProject[Iri('http://www.salsah.org/version/2.0/SwissBritNet')].add(AdminPermission.ADMIN_LISTS)
        user.update()
        user = User.read(con=self._connection, userId="volta", ignore_cache=True)
        self.assertEqual(user.inProject[Iri('oldap:HyperHamlet')], {AdminPermission.ADMIN_USERS, AdminPermission.ADMIN_RESOURCES})
        self.assertEqual(user.inProject[Iri('http://www.salsah.org/version/2.0/SwissBritNet')], {AdminPermission.ADMIN_MODEL, AdminPermission.ADMIN_LISTS})

    def test_update_user_in_project_modify_C(self):
        user = User(con=self._connection,
                    userIri=Iri("https://orcid.org/0000-0002-9991-2067"),
                    userId=Xsd_NCName("voltaB"),
                    familyName="VoltaB",
                    givenName="AlessandroB",
                    email="alessandro.volta@tension.com",
                    credentials="ElectricTensionB",
                    inProject={Iri('oldap:HyperHamlet'): {AdminPermission.ADMIN_USERS,
                                                          AdminPermission.ADMIN_RESOURCES},
                               Iri('http://www.salsah.org/version/2.0/SwissBritNet'): {AdminPermission.ADMIN_MODEL}},
                    hasPermissions={Iri('oldap:GenericView')})
        user.create()
        user = User.read(con=self._connection, userId="voltaB", ignore_cache=True)
        user.inProject[Iri('http://www.salsah.org/version/2.0/SwissBritNet')] = {AdminPermission.ADMIN_LISTS}
        user.update()
        user = User.read(con=self._connection, userId="voltaB", ignore_cache=True)
        self.assertEqual(user.inProject[Iri('oldap:HyperHamlet')], {AdminPermission.ADMIN_USERS, AdminPermission.ADMIN_RESOURCES})
        self.assertEqual(user.inProject[Iri('http://www.salsah.org/version/2.0/SwissBritNet')], {AdminPermission.ADMIN_LISTS})

    def test_update_user_in_project_modify_D(self):
        user = User(con=self._connection,
                    userIri=Iri("https://orcid.org/0000-0002-9991-2067"),
                    userId=Xsd_NCName("curie"),
                    familyName="Curie",
                    givenName="Marie",
                    email="marie.curie@radioactivity.com",
                    credentials="Radiation",
                    inProject={Iri('oldap:HyperHamlet'): {AdminPermission.ADMIN_USERS,
                                                          AdminPermission.ADMIN_RESOURCES},
                               Iri('http://www.salsah.org/version/2.0/SwissBritNet'): {AdminPermission.ADMIN_MODEL}},
                    hasPermissions={Iri('oldap:GenericView')})
        user.create()
        user = User.read(con=self._connection, userId="curie", ignore_cache=True)
        user.inProject[Iri('oldap:HyperHamlet')].discard(AdminPermission.ADMIN_RESOURCES)
        user.update()
        user = User.read(con=self._connection, userId="curie", ignore_cache=True)
        self.assertEqual(user.inProject[Iri('oldap:HyperHamlet')], {AdminPermission.ADMIN_USERS})
        self.assertEqual(user.inProject[Iri('http://www.salsah.org/version/2.0/SwissBritNet')], {AdminPermission.ADMIN_MODEL})

    def test_update_user_permissions_C(self):
        user = User(con=self._connection,
                    userIri=Iri("https://orcid.org/0000-0002-9991-2066"),
                    userId=Xsd_NCName("tesla"),
                    familyName="Tesla",
                    givenName="Nikolai",
                    email="nikolai.tesla@tesla.com",
                    credentials="Alternative current",
                    inProject={Iri('oldap:HyperHamlet'): {AdminPermission.ADMIN_USERS,
                                                          AdminPermission.ADMIN_RESOURCES}},
                    hasPermissions={Iri('oldap:GenericView')})
        user.create()
        user = User.read(con=self._connection, userId="tesla", ignore_cache=True)
        user.inProject[Iri('oldap:HyperHamlet')].add(AdminPermission.ADMIN_CREATE)
        user.update()
        user = User.read(con=self._connection, userId="tesla", ignore_cache=True)
        self.assertEqual(user.inProject[Iri('oldap:HyperHamlet')], {AdminPermission.ADMIN_USERS, AdminPermission.ADMIN_RESOURCES, AdminPermission.ADMIN_CREATE})


    #  #unittest.skip('Work in progress')
    def test_update_user_unpriv(self):
        user = User.read(con=self._unpriv, userId="bugsbunny", ignore_cache=True)
        user.credentials = "ChangedPassword"
        with self.assertRaises(OldapErrorNoPermission) as ex:
            user.update()
        self.assertEqual('Actor has no ADMIN_USERS permission for project oldap:HyperHamlet', str(ex.exception))

    #  #unittest.skip('Work in progress')
    def test_update_user_change_in_project(self):
        user = User(con=self._connection,
                    userIri=Iri("https://orcid.org/0000-0001-5925-2956"),
                    userId=Xsd_NCName("chiquet"),
                    familyName="Chiquet",
                    givenName="Vera",
                    email="vera.chiquet@unibas.ch",
                    credentials="Photography",
                    inProject={Iri('http://www.salsah.org/version/2.0/SwissBritNet'): {
                        AdminPermission.ADMIN_USERS,
                        AdminPermission.ADMIN_RESOURCES,
                        AdminPermission.ADMIN_CREATE}},
                    hasPermissions={Iri('oldap:GenericView')})
        user.create()
        del user
        user = User.read(con=self._connection, userId="chiquet", ignore_cache=True)
        user.inProject = InProjectClass({Iri('http://www.salsah.org/version/2.0/SwissBritNet'): {AdminPermission.ADMIN_OLDAP}})
        user.update()
        del user
        user = User.read(con=self._connection, userId="chiquet", ignore_cache=True)
        self.assertEqual(user.inProject, InProjectClass({Iri('http://www.salsah.org/version/2.0/SwissBritNet'): {AdminPermission.ADMIN_OLDAP}}))

    #  #unittest.skip('Work in progress')
    def test_update_user_empty_in_project(self):
        user = User(con=self._connection,
                    userIri=Iri("https://orcid.org/0000-0002-2553-8814"),
                    userId=Xsd_NCName("bsimpson"),
                    familyName="Simpson",
                    givenName="Bart",
                    email="bart.simpson@springfield.com",
                    credentials="AtomicPower",
                    hasPermissions={Iri('oldap:GenericView')})
        user.create()
        del user
        user = User.read(con=self._connection, userId="bsimpson", ignore_cache=True)
        user.inProject = InProjectClass({Iri('http://www.salsah.org/version/2.0/SwissBritNet'): {
            AdminPermission.ADMIN_USERS, AdminPermission.ADMIN_RESOURCES
        }})
        user.update()
        del user
        user = User.read(con=self._connection, userId="bsimpson", ignore_cache=True)
        self.assertEqual(user.inProject, InProjectClass({Iri('http://www.salsah.org/version/2.0/SwissBritNet'): {
            AdminPermission.ADMIN_USERS, AdminPermission.ADMIN_RESOURCES
        }}))

    #  #unittest.skip('Work in progress')
    def test_update_user_rm_in_project(self):
        user = User(con=self._connection,
                    userIri=Iri("https://orcid.org/0000-0001-5925-2956"),
                    userId=Xsd_NCName("chiquet"),
                    familyName="Chiquet",
                    givenName="Vera",
                    email="vera.chiquet@unibas.ch",
                    credentials="Photography",
                    inProject={Iri('http://www.salsah.org/version/2.0/SwissBritNet'): {
                        AdminPermission.ADMIN_USERS,
                        AdminPermission.ADMIN_RESOURCES,
                        AdminPermission.ADMIN_CREATE}},
                    hasPermissions={Iri('oldap:GenericView')})
        user.create()
        del user
        user = User.read(con=self._connection, userId="chiquet", ignore_cache=True)
        user.inProject = None
        user.update()
        del user
        user = User.read(con=self._connection, userId="chiquet", ignore_cache=True)
        self.assertFalse(user.inProject)

    #  #unittest.skip('Work in progress')
    def test_update_user_del_in_project(self):
        user = User(con=self._connection,
                    userIri=Iri("https://orcid.org/0000-0001-5925-2956"),
                    userId=Xsd_NCName("chiquet"),
                    familyName="Chiquet",
                    givenName="Vera",
                    email="vera.chiquet@unibas.ch",
                    credentials="Photography",
                    inProject={Iri('http://www.salsah.org/version/2.0/SwissBritNet'): {
                        AdminPermission.ADMIN_USERS,
                        AdminPermission.ADMIN_RESOURCES,
                        AdminPermission.ADMIN_CREATE}},
                    hasPermissions={Iri('oldap:GenericView')})
        user.create()
        del user
        user = User.read(con=self._connection, userId="chiquet", ignore_cache=True)
        del user.inProject
        user.update()
        del user
        user = User.read(con=self._connection, userId="chiquet", ignore_cache=True)
        self.assertFalse(user.inProject)

    #  #unittest.skip('Work in progress')
    def test_update_user_add_to_project(self):
        user = User(con=self._connection,
                    userIri=Iri("https://orcid.org/0000-0002-2553-8814"),
                    userId=Xsd_NCName("bsimpson"),
                    familyName="Simpson",
                    givenName="Bart",
                    email="bart.simpson@springfield.com",
                    credentials="AtomicPower",
                    inProject={Iri('http://www.salsah.org/version/2.0/SwissBritNet'): {
                        AdminPermission.ADMIN_USERS,
                        AdminPermission.ADMIN_RESOURCES,
                        AdminPermission.ADMIN_CREATE}},
                    hasPermissions={Iri('oldap:GenericView')})
        user.create()
        del user
        user = User.read(con=self._connection, userId="bsimpson", ignore_cache=True)
        user.inProject[Iri('oldap:HyperHamlet')] = {AdminPermission.ADMIN_RESOURCES, AdminPermission.ADMIN_CREATE}
        user.update()
        del user
        user = User.read(con=self._connection, userId="bsimpson", ignore_cache=True)
        self.assertEqual(user.inProject, InProjectClass(
            {
                Iri('http://www.salsah.org/version/2.0/SwissBritNet'): {
                    AdminPermission.ADMIN_USERS, AdminPermission.ADMIN_RESOURCES, AdminPermission.ADMIN_CREATE
                },
                Iri('oldap:HyperHamlet'): {
                    AdminPermission.ADMIN_RESOURCES, AdminPermission.ADMIN_CREATE
                }
            }
        ))

    #@unittest.skip('Work in progress')
    def test_update_user_rm_from_project(self):
        user = User(con=self._connection,
                    userIri=Iri("https://orcid.org/0000-0002-2553-8814"),
                    userId=Xsd_NCName("bsimpson"),
                    familyName="Simpson",
                    givenName="Bart",
                    email="bart.simpson@springfield.com",
                    credentials="AtomicPower",
                    inProject={
                        Iri('http://www.salsah.org/version/2.0/SwissBritNet'): {
                            AdminPermission.ADMIN_USERS,
                            AdminPermission.ADMIN_RESOURCES,
                            AdminPermission.ADMIN_CREATE
                        },
                        Iri('oldap:HyperHamlet'): {
                            AdminPermission.ADMIN_RESOURCES, AdminPermission.ADMIN_CREATE
                        }
                    },
                    hasPermissions={Iri('oldap:GenericView')})
        user.create()
        user = User.read(con=self._connection, userId="bsimpson", ignore_cache=True)
        user.inProject[Iri('oldap:HyperHamlet')] = None
        user.update()
        user = User.read(con=self._connection, userId="bsimpson", ignore_cache=True)
        self.assertEqual(user.inProject.get(Iri('oldap:HyperHamlet')), {})
        self.assertEqual(user.inProject[Iri('http://www.salsah.org/version/2.0/SwissBritNet')], {
                    AdminPermission.ADMIN_USERS, AdminPermission.ADMIN_RESOURCES, AdminPermission.ADMIN_CREATE
            }
        )

    #  #unittest.skip('Work in progress')
    def test_update_user_del_from_project(self):
        user = User(con=self._connection,
                    userIri=Iri("https://orcid.org/0000-0002-2553-8814"),
                    userId=Xsd_NCName("bsimpson"),
                    familyName="Simpson",
                    givenName="Bart",
                    email="bart.simpson@springfield.com",
                    credentials="AtomicPower",
                    inProject={Iri('http://www.salsah.org/version/2.0/SwissBritNet'): {
                        AdminPermission.ADMIN_USERS,
                        AdminPermission.ADMIN_RESOURCES,
                        AdminPermission.ADMIN_CREATE},
                        Iri('oldap:HyperHamlet'): {
                            AdminPermission.ADMIN_RESOURCES, AdminPermission.ADMIN_CREATE
                        }
                    },
                    hasPermissions={Iri('oldap:GenericView')})
        user.create()
        del user
        user = User.read(con=self._connection, userId="bsimpson", ignore_cache=True)
        del user.inProject[Iri('oldap:HyperHamlet')]
        user.update()
        del user
        user = User.read(con=self._connection, userId="bsimpson", ignore_cache=True)
        self.assertEqual(user.inProject, InProjectClass(
            {
                Iri('http://www.salsah.org/version/2.0/SwissBritNet'): {
                    AdminPermission.ADMIN_USERS, AdminPermission.ADMIN_RESOURCES, AdminPermission.ADMIN_CREATE
                }
            }
        ))

    #  #unittest.skip('Work in progress')
    def test_update_user_del_has_permissions(self):
        user = User(con=self._connection,
                    userIri=Iri("https://orcid.org/0000-0001-5925-2956"),
                    userId=Xsd_NCName("chiquet"),
                    familyName="Chiquet",
                    givenName="Vera",
                    email="vera.chiquet@unibas.ch",
                    credentials="Photography",
                    inProject={Iri('http://www.salsah.org/version/2.0/SwissBritNet'): {
                        AdminPermission.ADMIN_USERS,
                        AdminPermission.ADMIN_RESOURCES,
                        AdminPermission.ADMIN_CREATE}},
                    hasPermissions={Iri('oldap:GenericView')})
        user.create()
        del user
        user = User.read(con=self._connection, userId="chiquet", ignore_cache=True)
        del user.hasPermissions
        user.update()
        del user
        user = User.read(con=self._connection, userId="chiquet", ignore_cache=True)
        self.assertFalse(user.hasPermissions)

    #  #unittest.skip('Work in progress')
    def test_update_user_add_has_permissions(self):
        user = User(con=self._connection,
                    userIri=Iri("https://orcid.org/0000-0001-5925-2956"),
                    userId=Xsd_NCName("chiquet"),
                    familyName="Chiquet",
                    givenName="Vera",
                    email="vera.chiquet@unibas.ch",
                    credentials="Photography",
                    inProject={Iri('http://www.salsah.org/version/2.0/SwissBritNet'): {
                        AdminPermission.ADMIN_USERS,
                        AdminPermission.ADMIN_RESOURCES,
                        AdminPermission.ADMIN_CREATE}},
                    hasPermissions={Iri('oldap:GenericView')})
        user.create()
        del user
        user = User.read(con=self._connection, userId="chiquet", ignore_cache=True)
        user.hasPermissions.add(Iri('hyha:HyperHamletMember'))
        user.update()
        del user
        user = User.read(con=self._connection, userId="chiquet", ignore_cache=True)
        self.assertEqual(user.hasPermissions, {Iri('oldap:GenericView'), Iri('hyha:HyperHamletMember')})

    #  #unittest.skip('Work in progress')
    def test_update_user_add_has_permissions(self):
        user = User(con=self._connection,
                    userIri=Iri("https://orcid.org/0000-0001-5925-2956"),
                    userId=Xsd_NCName("chiquet"),
                    familyName="Chiquet",
                    givenName="Vera",
                    email="vera.chiquet@unibas.ch",
                    credentials="Photography",
                    inProject={Iri('http://www.salsah.org/version/2.0/SwissBritNet'): {
                        AdminPermission.ADMIN_USERS,
                        AdminPermission.ADMIN_RESOURCES,
                        AdminPermission.ADMIN_CREATE}})
        user.create()
        del user
        user = User.read(con=self._connection, userId="chiquet", ignore_cache=True)
        user.hasPermissions = {Iri('oldap:GenericView'), Iri('hyha:HyperHamletMember')}
        user.update()
        del user
        user = User.read(con=self._connection, userId="chiquet", ignore_cache=True)
        self.assertEqual(user.hasPermissions, {Iri('oldap:GenericView'), Iri('hyha:HyperHamletMember')})

    #  #unittest.skip('Work in progress')
    def test_update_user_add_bad_has_permissions(self):
        user = User(con=self._connection,
                    userIri=Iri("https://orcid.org/0000-0001-5925-2956"),
                    userId=Xsd_NCName("chiquet"),
                    familyName="Chiquet",
                    givenName="Vera",
                    email="vera.chiquet@unibas.ch",
                    credentials="Photography",
                    inProject={Iri('http://www.salsah.org/version/2.0/SwissBritNet'): {
                        AdminPermission.ADMIN_USERS,
                        AdminPermission.ADMIN_RESOURCES,
                        AdminPermission.ADMIN_CREATE}})
        user.create()
        del user
        user = User.read(con=self._connection, userId="chiquet", ignore_cache=True)
        user.hasPermissions = {Iri('oldap:GAGA'), Iri('hyha:HyperHamletMember')}
        with self.assertRaises(OldapErrorValue) as err:
            user.update()

    #  #unittest.skip('Work in progress')
    def test_update_user_rm_has_permissions(self):
        user = User(con=self._connection,
                    userIri=Iri("https://orcid.org/0000-0001-5925-2956"),
                    userId=Xsd_NCName("chiquet"),
                    familyName="Chiquet",
                    givenName="Vera",
                    email="vera.chiquet@unibas.ch",
                    credentials="Photography",
                    inProject={Iri('http://www.salsah.org/version/2.0/SwissBritNet'): {
                        AdminPermission.ADMIN_USERS,
                        AdminPermission.ADMIN_RESOURCES,
                        AdminPermission.ADMIN_CREATE}},
                    hasPermissions={Iri('oldap:GenericView'), Iri('hyha:HyperHamletMember')})
        user.create()
        del user
        user = User.read(con=self._connection, userId="chiquet", ignore_cache=True)
        user.hasPermissions.discard(Iri('hyha:HyperHamletMember'))
        user.update()
        del user
        user = User.read(con=self._connection, userId="chiquet", ignore_cache=True)
        self.assertEqual(user.hasPermissions, {Iri('oldap:GenericView')})

    #  unittest.skip('Work in progress')
    def test_update_user_unexisting_has_permissions(self):
        user = User(con=self._connection,
                    userIri=Iri("https://orcid.org/0000-0001-5925-2956"),
                    userId=Xsd_NCName("chiquet"),
                    familyName="Chiquet",
                    givenName="Vera",
                    email="vera.chiquet@unibas.ch",
                    credentials="Photography",
                    inProject={Iri('http://www.salsah.org/version/2.0/SwissBritNet'): {
                        AdminPermission.ADMIN_USERS,
                        AdminPermission.ADMIN_RESOURCES,
                        AdminPermission.ADMIN_CREATE}},
                    hasPermissions={Iri('oldap:GenericView'), Iri('hyha:HyperHamletMember')})
        user.create()
        del user
        user = User.read(con=self._connection, userId="chiquet", ignore_cache=True)
        user.hasPermissions.discard(Iri('oldap:GenericRestricted'))
        user.update()
        del user
        user = User.read(con=self._connection, userId="chiquet", ignore_cache=True)
        self.assertEqual(user.hasPermissions, {Iri('oldap:GenericView'), Iri('hyha:HyperHamletMember')})

    def test_user_update_active(self):
        user = User(con=self._connection,
                    userIri=Iri("https://orcid.org/0000-0003-4545-3559"),
                    userId=Xsd_NCName("jrosenthal"),
                    familyName="Rosenthal",
                    givenName="Joachim",
                    email="rosenthal@math.uzh.ch",
                    credentials="CryptoGraphy0*0@",
                    inProject={Iri('http://www.salsah.org/version/2.0/SwissBritNet'): {
                        AdminPermission.ADMIN_USERS,
                        AdminPermission.ADMIN_RESOURCES,
                        AdminPermission.ADMIN_CREATE}},
                    isActive=False,
                    hasPermissions={Iri('oldap:GenericView'), Iri('hyha:HyperHamletMember')})
        user.create()
        self.assertFalse(user.isActive)
        del user
        user = User.read(con=self._connection, userId="jrosenthal", ignore_cache=True)
        self.assertFalse(user.isActive)
        user.isActive = True
        self.assertIsInstance(user.isActive, Xsd_boolean)
        user.update()
        del user
        user = User.read(con=self._connection, userId="jrosenthal", ignore_cache=True)
        self.assertTrue(user.isActive)
        self.assertIsInstance(user.isActive, Xsd_boolean)

        user.isActive = False
        self.assertIsInstance(user.isActive, Xsd_boolean)
        user.update()
        del user
        user = User.read(con=self._connection, userId="jrosenthal", ignore_cache=True)
        self.assertFalse(user.isActive)
        self.assertIsInstance(user.isActive, Xsd_boolean)

    def test_user_password_change(self):
        user = User(con=self._connection,
                    userId=Xsd_NCName("speedy"),
                    familyName="Gonzales",
                    givenName="Speedy",
                    email="speedy.gonzales@mexico.com",
                    credentials="FastestMouseInMexico",
                    isActive=True)
        user.create()
        mycon = Connection(server='http://localhost:7200',
                           repo="oldap",
                           userId="speedy",
                           credentials="FastestMouseInMexico",
                           context_name="DEFAULT")
        user = User.read(con=mycon, userId="speedy", ignore_cache=True)
        user.credentials = "ElRatónMásRápidoDeMéxico"
        user.update()
        with self.assertRaises(OldapError) as err:
            mycon = Connection(server='http://localhost:7200',
                               repo="oldap",
                               userId="speedy",
                               credentials="FastestMouseInMexico",
                               context_name="DEFAULT")
        mycon = Connection(server='http://localhost:7200',
                           repo="oldap",
                           userId="speedy",
                           credentials="ElRatónMásRápidoDeMéxico",
                           context_name="DEFAULT")
        user = User.read(con=mycon, userId="speedy", ignore_cache=True)
        user.isActive = False
        with self.assertRaises(OldapErrorNoPermission) as err:
            user.update()

    def test_user_authorizations(self):
        user = User(con=self._unpriv,
                    userIri=Iri("https://orcid.org/0000-0001-9421-3434"),
                    userId=Xsd_NCName("niederer"),
                    familyName="Niederer",
                    givenName="Markus",
                    email="markus.niederer@gmail.com",
                    credentials="DendroChronologie",
                    inProject={Iri('http://www.salsah.org/version/2.0/SwissBritNet'): {
                        AdminPermission.ADMIN_USERS,
                        AdminPermission.ADMIN_RESOURCES,
                        AdminPermission.ADMIN_CREATE}},
                    hasPermissions={Iri('oldap:GenericView'), Iri('hyha:HyperHamletMember')})
        with self.assertRaises(OldapErrorNoPermission) as ex:
            user.create()

        user = User(con=self._connection,
                    userIri=Iri("https://orcid.org/0000-0001-9421-3434"),
                    userId=Xsd_NCName("niederer"),
                    familyName="Niederer",
                    email="markus.niederer@gmail.com",
                    givenName="Markus",
                    credentials="DendroChronologie",
                    inProject={Iri('http://www.salsah.org/version/2.0/SwissBritNet'): {
                        AdminPermission.ADMIN_USERS,
                        AdminPermission.ADMIN_RESOURCES,
                        AdminPermission.ADMIN_CREATE}},
                    hasPermissions={Iri('oldap:GenericView'), Iri('hyha:HyperHamletMember')})
        user.create()
        user = User.read(con=self._unpriv, userId="niederer", ignore_cache=True)
        user.familyName = "Niederer"
        user.inProject[Iri('http://www.salsah.org/version')] = {AdminPermission.ADMIN_CREATE}
        with self.assertRaises(OldapErrorNoPermission) as ex:
            user.update()
        with self.assertRaises(OldapErrorNoPermission) as ex:
            user.delete()


if __name__ == '__main__':
    unittest.main()



from copy import deepcopy
from enum import Enum
from typing import Set, Dict, Any, Self

from oldaplib.src.enums.action import Action
from oldaplib.src.enums.attributeclass import AttributeClass
from oldaplib.src.helpers.attributechange import AttributeChange
from oldaplib.src.helpers.context import Context
from oldaplib.src.xsd.iri import Iri
from oldaplib.src.xsd.xsd_boolean import Xsd_boolean
from oldaplib.src.xsd.xsd_qname import Xsd_QName
from oldaplib.src.xsd.xsd_datetime import Xsd_dateTime
from oldaplib.src.helpers.oldaperror import OldapError, OldapErrorNotFound, OldapErrorType, OldapErrorImmutable, OldapErrorValue
from oldaplib.src.helpers.query_processor import QueryProcessor
from oldaplib.src.iconnection import IConnection


#@strict
class Model:
    _con: IConnection
    _changed: Set[str]
    _creator: Iri | None
    _created: Xsd_dateTime | None
    _contributor: Iri | None
    _modified: Xsd_dateTime | None
    _attributes: dict[Enum, Any]
    _changeset: dict[AttributeClass, AttributeChange]

    def __init__(self, *,
                 connection: IConnection,
                 creator: Iri | None = None,
                 created: Xsd_dateTime | None = None,
                 contributor: Iri | None = None,
                 modified: Xsd_dateTime | None = None) -> None:
        # if not isinstance(connection, Connection):
        #     raise OldapError('"con"-parameter must be an instance of Connection')
        # if type(connection) != Connection:
        #     raise OldapError('"con"-parameter must be an instance of Connection')
        self._con = connection
        if not creator:
            creator = self._con.userIri
        if not contributor:
            contributor = self._con.userIri

        self._creator = creator
        self._created = created
        self._contributor = contributor
        self._modified = modified
        self._attributes = {}
        self._changeset = {}

    def __deepcopy__(self, memo: Dict[int, Any]) -> 'Model':
        cls = self.__class__
        instance = cls.__new__(cls)
        memo[id(self)] = instance
        instance._creator = deepcopy(self._creator, memo=memo)
        instance._created = deepcopy(self._created, memo=memo)
        instance._contributor = deepcopy(self._contributor, memo=memo)
        instance._modified = deepcopy(self._modified, memo=memo)
        instance._attributes = deepcopy(self._attributes, memo=memo)
        instance._changeset = {}

        return instance

    def check_consistency(self, attr: AttributeClass, value: Any) -> None:
        pass

    def pre_transform(self, attr: AttributeClass, value: Any) -> Any:
        return value

    def cleanup_setter(self, attr: AttributeClass, value: Any) -> None:
        pass

    def notifier(self, attr: AttributeClass, value: Any) -> None:
        pass

    def __str__(self) -> str:
        res = f'Creation: {self._created} by {self._creator}\n'
        res += f'Modified: {self._modified} by {self._contributor}\n'
        for attr, value in self._attributes.items():
            res += f'{attr} ({attr.value}): {value}\n'
        return res

    def __getitem__(self, attr: AttributeClass) -> Any:
        return self._attributes[attr]

    def get(self, attr: AttributeClass) -> Any:
        return self._attributes.get(attr)

    def __setitem__(self, attr: AttributeClass, value: Any) -> None:
        self._change_setter(attr, value)

    def __delitem__(self, attr: AttributeClass) -> None:
        if self._attributes.get(attr) is not None:
            self._changeset[attr] = AttributeChange(self._attributes[attr], Action.DELETE)
            del self._attributes[attr]

    def set_attributes(self, arguments: dict[str, Any], Attributes: type[Enum]) -> None:
        for name, value in arguments.items():
            if not isinstance(value, (bool, Xsd_boolean)) and not value:
                continue
            attr = Attributes.from_name(name)
            try:
                self._attributes[attr] = value if isinstance(value, attr.datatype) else attr.datatype(value)
            except ValueError as err:
                raise OldapErrorValue(err)
            if hasattr(self._attributes[attr], 'set_notifier'):
                self._attributes[attr].set_notifier(self.notifier, attr)
        for attr in Attributes:
            if attr.mandatory and not self._attributes.get(attr):
                raise OldapErrorType(f'Mandatory parameter {attr.name} is missing.')

    def _get_value(self: Self, attr: AttributeClass) -> Any | None:
        tmp = self._attributes.get(attr)
        if attr.datatype != Xsd_boolean and not tmp:
            return None
        return tmp

    def _set_value(self: Self, value: Any, attr: AttributeClass) -> None:
        self._change_setter(attr, value)

    def _del_value(self: Self, attr: AttributeClass) -> None:
        self._changeset[attr] = AttributeChange(self._attributes[attr], Action.DELETE)
        del self._attributes[attr]
        if hasattr(self, "notify"):
            self.notify()
        self.cleanup_setter(attr, None)

    def _change_setter(self, attr: AttributeClass, value: Any) -> None:
        if value is not None:
            if self._attributes.get(attr) == value:
                return
            if attr.immutable:
                raise OldapErrorImmutable(f'Attribute {attr.value} is immutable.')
            self.check_consistency(attr, value)
            value = self.pre_transform(attr, value)
        if self._attributes.get(attr) is None:
            if self._changeset.get(attr) is None:
                self._changeset[attr] = AttributeChange(None, Action.CREATE)
        else:
            if value is None:
                if self._changeset.get(attr) is None:
                    self._changeset[attr] = AttributeChange(deepcopy(self._attributes[attr]), Action.DELETE)
            else:
                if self._changeset.get(attr) is None:
                    self._changeset[attr] = AttributeChange(deepcopy(self._attributes[attr]), Action.REPLACE)
        if value is None:
            del self._attributes[attr]
        else:
            if not isinstance(value, attr.datatype):
                self._attributes[attr] = attr.datatype(value)
            else:
                self._attributes[attr] = value
            if hasattr(self._attributes[attr], 'set_notifier') and hasattr(self, 'notifier'):
                self._attributes[attr].set_notifier(self.notifier, attr)
        if hasattr(self, "notify"):
            self.notify()
        self.cleanup_setter(attr, value)

    @property
    def changeset(self) -> Dict[AttributeClass, AttributeChange]:
        """
        Return the changeset, that is dicst with information about all properties that have benn changed.
        This method is only for internal use or debugging...
        :return: A dictionary of all changes
        :rtype: Dict[ProjectAttr, ProjectAttrChange]
        """
        return self._changeset

    def clear_changeset(self) -> None:
        """
        Clear the changeset. This method is only for internal use or debugging...
        :return: None
        """
        for item in self._attributes:
            if hasattr(self._attributes[item], 'clear_changeset'):
                self._attributes[item].clear_changeset()
        self._changeset = {}

    @property
    def creator(self) -> Iri | None:
        """
        The creator of the project.
        :return: Iri of the creator of the project.
        :rtype: Iri | None
        """
        return self._creator

    @property
    def created(self) -> Xsd_dateTime | None:
        """
        The creation date of the project.
        :return: Creation date of the project.
        :rtype: Xsd_dateTime | None
        """
        return self._created

    @property
    def contributor(self) -> Iri | None:
        """
        The contributor of the project as Iri.
        :return: Iri of the contributor of the project.
        :rtype: Iri | None
        """
        return self._contributor

    @property
    def modified(self) -> Xsd_dateTime | None:
        """
        Modification date of the project.
        :return: Modification date of the project.
        :rtype: Xsd_dateTime | None
        """
        return self._modified


    def get_modified_by_iri(self, graph: Xsd_QName, iri: Iri) -> Xsd_dateTime:
        context = Context(name=self._con.context_name)
        sparql = context.sparql_context
        sparql += f"""
        SELECT ?modified
        FROM {graph}
        WHERE {{
            {iri.toRdf} dcterms:modified ?modified
        }}
        """
        jsonobj = None
        if self._con.in_transaction():
            jsonobj = self._con.transaction_query(sparql)
        else:
            jsonobj = self._con.query(sparql)
        res = QueryProcessor(context, jsonobj)
        if len(res) != 1:
            raise OldapErrorNotFound(f'No resource found with iri "{iri}".')
        for r in res:
            return r['modified']

    def set_modified_by_iri(self,
                            graph: Xsd_QName,
                            iri: Iri,
                            old_timestamp: Xsd_dateTime,
                            timestamp: Xsd_dateTime) -> None:
        try:
            context = Context(name=self._con.context_name)
            sparql = context.sparql_context
            sparql += f"""
            WITH {graph}
            DELETE {{
                ?res dcterms:modified {old_timestamp.toRdf} .
                ?res dcterms:contributor ?contributor .
            }}
            INSERT {{
                ?res dcterms:modified {timestamp.toRdf} .
                ?res dcterms:contributor {self._con.userIri.toRdf} .
            }}
            WHERE {{
                BIND({iri.toRdf} as ?res)
                ?res dcterms:modified {old_timestamp.toRdf} .
                ?res dcterms:contributor ?contributor .
            }}
            """
        except Exception as e:
            raise OldapError(e)
        if self._con.in_transaction():
            self._con.transaction_update(sparql)
        else:
            self._con.update_query(sparql)
